from typing import List

from ..tokenizer import GroupToken
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib


def lib_exit(ctx: TranspilerContext, _args: List[GroupToken], token: GroupToken):
    ctx.file.append("return " + token.value[5:-1])


add_lib(FunctionDeclaration(
    type="python-raw",
    name="exit",
    returns="void",
    function=lib_exit
))



from typing import Union, List

from ._base import CompileTimeValue
from .nbt import CplNBT, object_get_index_nbt, val_nbt
from .tuple import CplTuple
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import CplDefArray, get_uuid


class CplArrayNBT(CplNBT):
    def __init__(self, token: Union[Token, None], location: str, type: CplDefArray):
        super().__init__(token, location, type)
        self.unique_type = type

    def _set(self, ctx, cpl):
        if isinstance(cpl, CplTuple) and len(cpl.value) == 0:
            ctx.file.append(f"data modify {self.location} set value []")
            return self
        return super()._set(ctx, cpl)

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplString) and index.value == "length":
            eid = f"int_{get_uuid()} __temp__"
            ctx.file.append(f"execute store result score {eid} run data get {self.location}")
            return CplScore(self.token, eid)
        if isinstance(index, CplInt) or isinstance(index, CplString):
            ind = str(index.value)
            if not ind.isnumeric():
                return None
            content = self.unique_type.content
            return val_nbt(self.token, self.location + "[" + ind + "]", content)

        return object_get_index_nbt(ctx, self.unique_type.content, self, index)

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        if index == "pop":
            if len(arguments) != 0:
                raise_syntax_error("Expected 0 arguments for <array>.pop()", self.token)
            eid = f"storage temp _{get_uuid()}"
            ctx.file.append(f"data modify {eid} set from {self.location}[-1]")
            ctx.file.append(f"data remove {self.location}[-1]")
            return val_nbt(self.token, eid, self.unique_type.content)
        if index == "push":
            if len(arguments) == 0:
                raise_syntax_error("Expected at least 1 argument for array.push()", self.token)
            for arg in arguments:
                ctx.file.append(f"data modify {self.location} append {arg.get_data_str(ctx)}")
            return self.get_index(ctx, CplString(self.token, "length"))
        if index == "insert":
            if len(arguments) != 2:
                raise_syntax_error("Expected 2 arguments for <array>.insert()", self.token)
            a0 = arguments[0]
            a1 = arguments[1]
            if isinstance(a0, CplInt):
                ctx.file.append(f"data modify {self.location} insert {a0.value} {a1.get_data_str(ctx)}")
                return a1
            if isinstance(a0, CplIntNBT):
                fl = []
                ctx.file.append(f"$data modify {self.location} insert $(_0) {a1.get_data_str(ctx)}")
                file_name = ctx.transpiler.get_temp_file_name(fl)
                a0.cache(ctx, nbt_loc="storage array_insert _0", force="nbt")
                ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage array_insert")
                return a1
            raise_syntax_error("Expected an int as the first argument for array.insert()", self.token)
        if index == "merge":
            if len(arguments) != 1:
                raise_syntax_error("Expected 1 argument for array.merge()", self.token)
            a0 = arguments[0]
            if isinstance(a0, CplArray):
                a0 = a0.cache(ctx)
            if not isinstance(a0, CplArrayNBT):
                raise_syntax_error("Expected an array as the first argument for array.insert()", self.token)
            if a0.unique_type != self.unique_type:
                raise_syntax_error("Expected an array of the same type for array.merge()", self.token)
            ctx.file.append(f"data modify {self.location} merge {a0.get_data_str(ctx)}")
            return self
        return None

    def _add(self, ctx, cpl):
        if self.unique_type != cpl.unique_type:
            return None
        # TODO: array merging
        return None

    def _and(self, ctx, cpl):
        return cpl

    def _or(self, ctx, cpl):
        return self


from .array import CplArray
from .int import CplInt
from .nbtint import CplIntNBT
from .score import CplScore
from .string import CplString



from typing import Any, Dict, List

from .error import raise_syntax_error, raise_syntax_error_t

SET_OP = ["=", "+=", "-=", "*=", "/=", "%="]


def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


OPERATORS = list("+-/!=%")
OPERATORS_L = [
    "==",
    "!=",
    "++",
    "--",
    "+=",
    "-=",
    "*=",
    "/=",
    "%=",
    "**=",
    "**",
    "*",
    "&&",
    "||",
    ">=",
    ">",
    "<=",
    "<",
    "is not",
    "is",
    "and",
    "or",
]
WORD_OPERATORS = ["is not", "is", "and", "or"]
INC_OP = ["++", "--"]
CMP_OP = ["==", "!=", ">", "<", ">=", "<=", "is", "is not"]
CMP_COMBINE_OP = ["&&", "||", "and", "or"]
SYMBOL = list("}]){[(:,@\\.~^$")
QUOTES = list("\"'")
WHITESPACE = list(" \t\v")
EOL_CHAR = "\n"
EOE_CHAR = ";"
NON_WORD_CHARACTERS = OPERATORS + QUOTES + WHITESPACE + SYMBOL + [EOL_CHAR, EOE_CHAR]

KEYWORDS = [
    "if",
    "unless",
    "else",
    "elif",
    "for",
    "while",
    "until",
    "function",
    "return",
    "break",
    "continue",
    "define",
    # "int",
    # "float",
    # "string",
    "from",
    "import",
    "as",
]
EXTRA_KEYWORDS = ["and", "or", "is", "not"]

OPEN_GROUP = list("{[(")
CLOSE_GROUP = list("}])")
CLOSE_MATCH = {"}": "{", "]": "[", ")": "("}


class Token:
    def __init__(self, code: str, type: Any, start: int, end: int):
        self.code = code
        self.type = type
        self.start = start
        self.end = end
        self.value = ""
        self.update_value()

    def update_value(self):
        self.value = self.code[self.start: self.end]
        return self.value

    def __str__(self) -> str:
        return f"Token(type={self.type}, start={self.start}, end={self.end}, value={self.value})"


from .utils import (
    FLOAT_TYPE,
    INT_TYPE,
    STRING_TYPE,
    CplDef,
    CplDefArray,
    CplDefObject,
    TokenType, SELECTOR_TYPE, CplDefFunction,
)

EmptyToken = Token("", TokenType.SYMBOL, 0, 0)


class GroupToken(Token):
    def __init__(self, code: str, start: int, end: int, children=None):
        super().__init__(code, TokenType.GROUP, start, end)
        self.parent: GroupToken = EmptyGroup
        self.children: List[Token | GroupToken] = children or []
        self.open: Token = EmptyToken
        self.close: Token = EmptyToken
        self.func: Token | None = None
        self._temp: Any = None

    def cpl(self, ctx):
        return ctx.transpiler.tokens_to_cpl(ctx, self.children)


EmptyGroup: GroupToken = Token("", TokenType.GROUP, 0, 0)  # type: ignore


class SelectorIdentifierToken(Token):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            selector: Token,
            sep: Token,
            name: Token,
    ):
        super().__init__(code, TokenType.SELECTOR_IDENTIFIER, start, end)
        self.selector = selector
        self.sep = sep
        self.name = name


class BlockIdentifierToken(Token):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            block: GroupToken,
            sep: Token,
            name: Token,
    ):
        super().__init__(code, TokenType.BLOCK_IDENTIFIER, start, end)
        self.block = block
        self.sep = sep
        self.name = name


class LambdaFunctionToken(Token):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            arguments: GroupToken,
            body: GroupToken
    ):
        super().__init__(code, TokenType.LAMBDA_FUNCTION, start, end)
        self.arguments = arguments
        self.body = body


def cpl_def_from_tokens(classes, tokens):
    # type: (Dict[str, Any] | List[str], List[Token]) -> CplDef | None
    """
    Valid type examples:
    int
    float
    string
    int[]
    string[][][]
    {"a": int, "b": float}
    """
    if len(tokens) == 0:
        return None
    type = None
    t0 = tokens[0]
    if t0.value == "int":
        type = INT_TYPE
    if t0.value == "float":
        type = FLOAT_TYPE
    if t0.value == "string":
        type = STRING_TYPE
    if t0.value == "selector":
        type = SELECTOR_TYPE
    if isinstance(t0, GroupToken) and t0.open.value == "(" and len(t0.children) > 2:
        t0c = t0.children
        if not isinstance(t0c[0], GroupToken) or t0c[0].open.value != "(" or t0c[1].value != "=" or t0c[2].value != ">":
            return None

        type = CplDefFunction(
            list(map(lambda x: cpl_def_from_tokens(classes, x), split_tokens(t0c[0].children))),
            cpl_def_from_tokens(classes, t0c[3:])
        )
    if isinstance(t0, GroupToken) and t0.open.value == "{":
        dct = {}
        spl = split_tokens(t0.children)
        for part in spl:
            if (
                    len(part) < 3
                    or part[0].type not in {TokenType.STRING_LITERAL, TokenType.IDENTIFIER}
                    or part[1].value != ":"
            ):
                return None
            k = part[0].value
            if part[0].type == TokenType.STRING_LITERAL:
                k = k[1:-1]
            dct[k] = cpl_def_from_tokens(classes, part[2:])
        type = CplDefObject(dct, None)
    if not type:
        if t0.type == TokenType.IDENTIFIER:
            if t0.value not in classes:
                return None
            type = classes[t0.value].attributes.unique_type if isinstance(classes, dict) else CplDefObject({}, t0.value)
        else:
            return None
    for t in tokens[1:]:
        if not isinstance(t, GroupToken) or t.open.value != "[" or len(t.children) != 0:
            return None
        type = CplDefArray(type)
    return type


def split_tokens(tokens: List[Token], sep=",", comma_errors=True) -> List[List[Token]]:
    ind = 0
    parts = [[]]
    while True:
        if ind == len(tokens):
            break
        token = tokens[ind]
        if token.value == sep:
            if comma_errors and len(parts[-1]) == 0:
                raise_syntax_error(f"Unexpected '{sep}'", token)
            parts.append([])
        else:
            parts[-1].append(token)
        ind += 1

    if comma_errors and len(parts[-1]) == 0:
        if len(parts) != 1:
            raise_syntax_error(f"Unexpected '{sep}'", tokens[-1])
        parts.pop()

    return parts


# Convert {[(...)]} and func(...) into its own tokens
def group_tokens(tokens: List[Token]) -> List[Token]:
    length = len(tokens)
    if length == 0:
        return []
    code = tokens[0].code
    parent: GroupToken = GroupToken(code, 0, len(code))
    base = parent

    index = -1
    while True:
        index += 1
        if index >= length:
            break
        token = tokens[index]
        pc = parent.children
        lc = len(pc)
        if token.type == TokenType.SYMBOL and token.value in OPEN_GROUP:
            func = None

            # @r[...something]
            is_selector = (
                    token.value == "["
                    and index != 0
                    and lc != 0
                    and (
                            pc[-1].type == TokenType.SELECTOR
                            or pc[-1].type == TokenType.SELECTOR_IDENTIFIER
                    )
                    and len(pc[-1].selector.value
                            if isinstance(pc[-1], SelectorIdentifierToken)
                            else pc[-1].value) == 2  # has to be a short selector like @a, @r
            )

            # func(arguments...)
            if (
                    not is_selector
                    and token.value == "("
                    and index != 0
                    and lc != 0
                    and pc[-1].type == TokenType.IDENTIFIER
                    and pc[-1].value[0] not in list("0123456789.")
            ):
                func = tokens[index - 1]
                pc.pop()

            tok = GroupToken(code, token.start, 0)
            tok.parent = parent
            tok.children = []
            tok.open = token
            tok.func = None

            if is_selector:
                tok.type = (
                    TokenType.SELECTOR
                    if pc[-1].type == TokenType.SELECTOR
                    else TokenType.SELECTOR_IDENTIFIER
                )

            if func is not None:
                tok.type = TokenType.FUNCTION_CALL
                tok.func = func
                tok.start = func.start

            is_block_identifier = (
                    token.value == "["
                    and len(pc) > 1
                    and pc[-1].value == ":"
                    and pc[-2].type == TokenType.IDENTIFIER
            )

            if is_block_identifier:
                tok._temp = "block"

            if not is_selector and not is_block_identifier:
                pc.append(tok)
            parent = tok
            continue
        if token.type == TokenType.SYMBOL and token.value in CLOSE_GROUP:
            if parent.parent is None or parent.open.value != CLOSE_MATCH[token.value]:
                raise_syntax_error("Unexpected parentheses", token)
            parent.close = token
            parent.end = token.end
            parent.update_value()

            ppc = parent.parent.children

            if (token.value == "}" and len(ppc) > 3 and ppc[-2].value == ">" and ppc[-3].value == "="
                    and ppc[-4].type == TokenType.GROUP and ppc[-4].open.value == "("):
                ppc.pop()
                ppc.pop()
                ppc.pop()
                arguments = ppc.pop()
                ppc.append(
                    LambdaFunctionToken(
                        parent.code,
                        arguments.start,
                        parent.end,
                        arguments,
                        parent
                    )
                )

            if parent._temp == "block":
                colon = ppc.pop()
                ident = ppc.pop()
                ppc.append(
                    BlockIdentifierToken(
                        parent.code,
                        ident.start,
                        parent.end,
                        parent,
                        colon,
                        ident,
                    )
                )

            if (
                    parent.type == TokenType.SELECTOR
                    or parent.type == TokenType.SELECTOR_IDENTIFIER
            ):
                merge = ppc[-1]
                merge.end = parent.end
                merge.update_value()
                if isinstance(merge, SelectorIdentifierToken):
                    merge.selector.end = parent.end
                    merge.selector.update_value()
            parent = parent.parent
            continue
        pc.append(token)

    if base != parent:
        parent.end = parent.start + 1
        raise_syntax_error("Incomplete parentheses", parent)

    return parent.children


def _tokenize_iterate(
        code: str, tokens: List[Token], macros: List, index: List[int], can_macro: bool
):
    char = code[index[0]]
    if char == EOE_CHAR:
        tokens.append(Token(code, TokenType.EOE, index[0], index[0] + 1))
        return
    if char in WHITESPACE:
        # tokens.append(Token(code, TokenType.WHITESPACE, index[0], index[0] + 1))
        return
    if char == EOL_CHAR:
        tokens.append(Token(code, TokenType.EOL, index[0], index[0] + 1))
        return
    if (
            char == "@"
            and index[0] != len(code) - 1
            and code[index[0] + 1] in list("praens")
    ):
        selToken = Token(code, TokenType.SELECTOR, index[0], index[0] + 2)
        if (
                len(tokens) > 1
                and tokens[-1].value == ":"
                and tokens[-1].end == index[0]
                and tokens[-2].type == TokenType.IDENTIFIER
        ):
            token = SelectorIdentifierToken(
                code,
                start=tokens[-2].start,
                end=index[0] + 2,
                selector=selToken,
                sep=tokens[-1],
                name=tokens[-2],
            )
            tokens.pop()
            tokens.pop()
            tokens.append(token)
            index[0] += 1
            return
        tokens.append(selToken)
        index[0] += 1
        return
    if code[index[0]: index[0] + 8] == "#define " and (
            len(tokens) == 0 or tokens[-1].type == TokenType.EOL
    ):
        if not can_macro:
            raise_syntax_error_t("Unexpected #define", code, index[0], index[0] + 8)
        si = index[0] + 8
        while index[0] < len(code):
            if code[index[0]] == "\n":
                break
            index[0] += 1
        index[0] -= 1
        m = code[si: index[0] + 1].split(" ")
        macros.append([m[0], _tokenize(" ".join(m[1:]), False)[0][:-1]])
        return
    if char == ":" and index[0] != len(code) - 1 and code[index[0] + 1] == ":":
        tokens.append(Token(code, TokenType.SYMBOL, index[0], index[0] + 2))
        index[0] += 1
        return
    if char in SYMBOL:
        tokens.append(Token(code, TokenType.SYMBOL, index[0], index[0] + 1))
        return
    if char == "/" and index[0] != len(code) - 1 and code[index[0] + 1] == "/":
        while index[0] < len(code):
            if code[index[0]] == "\n":
                index[0] -= 1
                break
            index[0] += 1
        return
    if char == "/" and index[0] != len(code) - 1 and code[index[0] + 1] == "*":
        line_index = 0
        while index[0] < len(code):
            if code[index[0]] == "\n":
                line_index = index[0]
            if (
                    code[index[0]] == "*"
                    and index[0] + 1 < len(code)
                    and code[index[0] + 1] == "/"
            ):
                index[0] += 1
                if line_index != 0:
                    tokens.append(
                        Token(code, TokenType.EOL, line_index, line_index + 1)
                    )
                break
            index[0] += 1
        return

    for op in OPERATORS_L:
        op_got = code[index[0]: index[0] + len(op)]
        if op_got == op and (op not in WORD_OPERATORS or code[index[0] + len(op)] == " "):
            tokens.append(Token(code, TokenType.OPERATOR, index[0], index[0] + len(op)))
            index[0] += len(op) - 1
            return
    if char in OPERATORS:
        tokens.append(Token(code, TokenType.OPERATOR, index[0], index[0] + 1))
        return
    if char in QUOTES:
        start_index = index[0]
        index[0] += 1
        backslash = False
        while index[0] < len(code):
            char2 = code[index[0]]
            if (char2 == char and not backslash) or char2 == EOL_CHAR:
                break
            if char == "\\":
                backslash = not backslash
            else:
                backslash = False
            index[0] += 1
        if index[0] == len(code) or code[index[0]] == EOL_CHAR:
            raise_syntax_error_t(
                "Unterminated string", code, start_index, start_index + 1
            )
        value = code[start_index + 1: index[0]]
        token: Token = Token(code, TokenType.STRING_LITERAL, start_index, index[0] + 1)
        tokens.append(token)
        return
    start_index = index[0]
    index[0] += 1
    while index[0] < len(code):
        char2 = code[index[0]]
        if char2 in NON_WORD_CHARACTERS:
            index[0] -= 1
            break
        index[0] += 1
    if index[0] == len(code):
        index[0] -= 1
    value = code[start_index: index[0] + 1]
    type = TokenType.IDENTIFIER
    if value in KEYWORDS:
        type = TokenType.KEYWORD
    elif str.isnumeric(value):
        type = TokenType.INT_LITERAL
    if type == TokenType.INT_LITERAL and len(tokens) > 0 and tokens[-1].value == "-" and (
            len(tokens) == 1 or tokens[-2].type in {TokenType.OPERATOR, TokenType.SYMBOL}):
        tokens.pop()
        start_index -= 1
    if (
            type == TokenType.INT_LITERAL
            and len(tokens) > 0
            and tokens[-1].value == "."
            and tokens[-1].end == start_index
            and (len(tokens) == 1 or tokens[-2].end != start_index - 1 or tokens[-2].type == TokenType.INT_LITERAL)):
        if len(tokens) > 1 and tokens[-2].type == TokenType.INT_LITERAL:
            tokens.pop()
        start = tokens.pop().start
        tokens.append(Token(code, TokenType.FLOAT_LITERAL, start, index[0] + 1))
        return True
    new_t = Token(code, type, start_index, index[0] + 1)
    if (type == TokenType.IDENTIFIER and len(tokens) > 1 and tokens[-1].value == ":"
            and tokens[-2].type == TokenType.IDENTIFIER):
        sep = tokens.pop()
        ident = tokens.pop()
        tokens.append(SelectorIdentifierToken(code, ident.start, index[0] + 1, new_t, sep, ident))
        return True
    tokens.append(new_t)


def _tokenize(code: str, can_macro=True):
    code = code.replace("\r", "")

    tokens: List[Token] = []
    macros = []

    index = [0]
    length = len(code)
    while index[0] < length:
        _tokenize_iterate(code, tokens, macros, index, can_macro)
        index[0] += 1

    tokens.append(Token(code, TokenType.EOF, length, length))
    return tokens, macros


def apply_macros(tokens: List[Token], macros: List):
    new_tokens = []
    for token in tokens:
        found = False
        for macro in macros:
            if token.type == TokenType.IDENTIFIER and token.value == macro[0]:
                found = True
                new_tokens.extend(macro[1])
                break
        if not found:
            new_tokens.append(token)
    return new_tokens


def tokenize(code: str):
    (tokens, macros) = _tokenize(code)
    new_tokens = group_tokens(apply_macros(tokens, macros))
    return new_tokens, macros



import math
from typing import List, Any

from ..cpl._base import CompileTimeValue
from ..cpl.float import CplFloat
from ..cpl.int import CplInt
from ..cpl.nbt import CplNBT
from ..cpl.nbtfloat import CplFloatNBT
from ..cpl.nbtint import CplIntNBT
from ..cpl.score import CplScore
from ..error import raise_syntax_error
from ..transpiler import TranspilerContext, add_lib, CustomCplObject
from ..utils import FLOAT_PREC, VariableDeclaration, get_uuid


def helper_float0_check(ctx: TranspilerContext, args: List[CompileTimeValue], token, method: str):
    if len(args) != 1:
        raise_syntax_error(f"Expected 1 argument for {method}()", token)
    n = args[0]
    if n.unique_type.type not in {"int", "float"}:
        raise_syntax_error(f"Expected an int or float argument for {method}()", token)


def helper_float0(ctx: TranspilerContext, args: List[CompileTimeValue], token, method: str,
                  score_loc=None) -> float | CplScore:
    helper_float0_check(ctx, args, token, method)
    n = args[0]
    if isinstance(n, CplInt) or isinstance(n, CplFloat):
        return float(n.value)
    return n.cache(ctx, score_loc=score_loc, force="score", force_t="float")


def lib_sqrt(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    tr = ctx.transpiler
    helper = helper_float0(ctx, args, token, "Math.sqrt", "__sqrt__x __temp__")
    if isinstance(helper, float):
        return CplFloat(token, math.sqrt(helper))

    if "has_sqrt_init" not in tr.data:
        tr.data["has_sqrt_init"] = True
        tr.files["__load__"].append("scoreboard players set __sqrt__2 __temp__ 2")

    # y^2 - x = 0
    # y -= (y^2 - x)/(2 y)

    # y -= (y - x/y) / 2

    # z = y
    # w = x
    # w /= y
    # z -= w
    # z /= 2

    # y -= z

    last_y = CplScore(token, "__sqrt__last_output __temp__", "float")
    y = CplScore(token, "__sqrt__output __temp__", "float")
    z = CplScore(token, "__sqrt__z __temp__", "float")
    w = CplScore(token, "__sqrt__w __temp__", "float")

    # initial guess: y = x
    y._set(ctx, helper)

    if "__lib__/sqrt_loop" not in tr.files:
        fil = [f"scoreboard players operation {last_y.location} = {y.location}"]
        tr.files["__lib__/sqrt_loop"] = fil
        loop_ctx = ctx.clone_at("__lib__/sqrt_loop")
        z._set(loop_ctx, y)
        w._set(loop_ctx, helper)
        w._set_div(loop_ctx, y)
        z._set_sub(loop_ctx, w)
        z._set_div(loop_ctx, CplScore(token, "__sqrt__2 __temp__", "int"))
        y._set_sub(loop_ctx, z)
        fil.append(
            f"execute unless score {last_y.location} = {y.location} run function {tr.pack_namespace}:__lib__/sqrt_loop")

    ctx.file.append(f"function {tr.pack_namespace}:__lib__/sqrt_loop")

    return CplScore(token, "__sqrt__output __temp__", "float")


def lib_isqrt(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    if len(args) != 1:
        raise_syntax_error("Expected 1 argument for Math.isqrt()", token)
    n = args[0]
    if n.unique_type.type != "int":
        raise_syntax_error("Expected an int argument for Math.isqrt()", token)
    if isinstance(n, CplInt):
        return CplInt(token, math.isqrt(n.value))
    n.cache(ctx, score_loc="__isqrt__x", force="score", force_t="int")

    tr = ctx.transpiler
    file = ctx.file
    if "has_sqrt_init" not in tr.data:
        tr.data["has_sqrt_init"] = True
        tr.files["__load__"].append("scoreboard players set __sqrt__2 __temp__ 2")
        tr.files["__load__"].append("scoreboard players set __sqrt__4 __temp__ 4")

    file.extend(
        [
            "scoreboard players operation __isqrt__x_4 __temp__ = __isqrt__x __temp__",
            "scoreboard players operation __isqrt__x_4 __temp__ /= __isqrt__4 __temp__",
            "scoreboard players operation __isqrt__output __temp__ = __isqrt__x_4 __temp__",
            f"function {tr.pack_namespace}:__lib__/__isqrt__loop",
        ]
    )

    tr.files["__lib__/isqrt_loop"] = [
        "scoreboard players operation __isqrt__last_output __temp__ = __isqrt__output __temp__",
        "scoreboard players operation __isqrt__output __temp__ /= __isqrt__2 __temp__",
        "scoreboard players operation __isqrt__x_t __temp__ =  __isqrt__x_4 __temp__",
        "scoreboard players operation __isqrt__x_t __temp__ /= __isqrt__output __temp__",
        "scoreboard players operation __isqrt__output __temp__ += __isqrt__x_t __temp__",
        "execute unless score __isqrt__last_output __temp__ = __isqrt__output __temp__ run function $PACK_NAME$:__lib__/__isqrt__loop",
    ]

    return CplScore(token, "__isqrt__output __temp__", "int")


def lib_cbrt(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    tr = ctx.transpiler
    helper = helper_float0(ctx, args, token, "Math.cbrt", "__cbrt__x __temp__")
    if isinstance(helper, float):
        return CplFloat(token, helper ** (1 / 3))

    if "has_cbrt_init" not in tr.data:
        tr.data["has_cbrt_init"] = True
        tr.files["__load__"].append(
            f"scoreboard players set __cbrt__3d2 __temp__ {int(FLOAT_PREC * 3 / 2)}"
        )
        tr.files["__load__"].append(
            f"scoreboard players set __cbrt__4d3 __temp__ {int(FLOAT_PREC * 4 / 3)}"
        )

    ctx.file.extend(
        [
            "scoreboard players operation __cbrt__x __temp__ *= __cbrt__4d3 __temp__",
            "scoreboard players operation __cbrt__output __temp__ = __cbrt__x __temp__",
            f"function {tr.pack_namespace}:__lib__/__cbrt__loop",
        ]
    )

    tr.files["__lib__/cbrt_loop"] = [
        "scoreboard players operation __cbrt__last_output __temp__ = __cbrt__output __temp__",
        "scoreboard players operation __cbrt__output __temp__ *= __cbrt__3d2 __temp__",
        "scoreboard players operation __cbrt__output __temp__ /= FLOAT_PREC __temp__",
        "scoreboard players operation __cbrt__x_t __temp__ =  __cbrt__x __temp__",
        "scoreboard players operation __cbrt__x_t __temp__ *= FLOAT_PREC __temp__",
        "scoreboard players operation __cbrt__x_t __temp__ /= __cbrt__output __temp__",
        "scoreboard players operation __cbrt__x_t __temp__ *= FLOAT_PREC __temp__",
        "scoreboard players operation __cbrt__x_t __temp__ /= __cbrt__output __temp__",
        "scoreboard players operation __cbrt__output __temp__ += __cbrt__x_t __temp__",
        "execute unless score __cbrt__last_output __temp__ = __cbrt__output __temp__ run function $PACK_NAME$:__lib__/__cbrt__loop",
    ]

    return CplScore(token, "__cbrt__output __temp__", "float")


def lib_floor(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper = helper_float0(ctx, args, token, "Math.floor")
    if isinstance(helper, float):
        return CplFloat(token, math.floor(helper))
    return helper


def lib_ceil(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper = helper_float0(ctx, args, token, "Math.ceil", "__ceil__x __temp__")
    if isinstance(helper, float):
        return CplFloat(token, math.ceil(helper))
    n = args[0]
    if n.unique_type.type == "int":
        return n

    ctx.file.append(f"scoreboard players add __ceil__x __temp__ {int(FLOAT_PREC - 1)}")
    ctx.file.append(f"scoreboard players operation __ceil__x __temp__ /= FLOAT_PREC __temp__")
    return helper


def lib_round(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper = helper_float0(ctx, args, token, "Math.round", "__round__x __temp__")
    if isinstance(helper, float):
        return CplFloat(token, round(helper))
    n = args[0]
    if n.unique_type.type == "int":
        return n

    ctx.file.append(f"scoreboard players add __round__x __temp__ {int(FLOAT_PREC / 2)}")
    ctx.file.append(f"scoreboard players operation __round__x __temp__ /= FLOAT_PREC __temp__")
    return helper


def _min_max_help(x, y, func_name):
    return x.value < y.value if func_name == "min" else x.value > y.value


def lib_min_max(
        ctx: TranspilerContext, arguments: List[CompileTimeValue], token, func_name
):
    if len(arguments) == 0:
        raise_syntax_error(f"Expected at least 1 argument for Math.{func_name}()", token)
    min_num: Any = None
    result_type = "int"
    non_literals = []
    for arg in arguments:
        if arg.unique_type.type not in {"int", "float"}:
            raise_syntax_error(f"Expected an int or float argument for Math.{func_name}()", token)
        if arg.unique_type.type == "float":
            result_type = "float"
        if isinstance(arg, CplInt) or isinstance(arg, CplFloat):
            if not min_num or _min_max_help(arg, min_num, func_name):
                min_num = arg
        else:
            non_literals.append(arg)

    result = min_num or CplInt(token, 0)
    for arg in non_literals:
        op = "<" if func_name == "min" else ">"
        if not isinstance(arg, CplScore):
            arg = arg.cache(ctx, force="score", force_t=result_type)
        if isinstance(result, CplInt) or isinstance(result, CplFloat):
            v = result.value if isinstance(result, CplInt) else int(result.value * FLOAT_PREC)
            if func_name == "min":
                cmp = f"..{v - 1}"
            else:
                cmp = f"{v + 1}.."
            ctx.file.append(f"execute if score {arg.location} matches {cmp} "
                            f"run scoreboard players set __min__output __temp__ {v}")
            result = CplScore(token, "__min__output __temp__", result_type)
        else:
            ctx.file.append(f"execute if score {arg.location} {op} {result.location} "
                            f"run scoreboard players set __min__output __temp__ {result.location}")
            result = CplScore(token, "__min__output __temp__", result_type)

    return result


def lib_min(ctx: TranspilerContext, arguments: List[CompileTimeValue], token):
    return lib_min_max(ctx, arguments, token, "min")


def lib_max(ctx: TranspilerContext, arguments: List[CompileTimeValue], token):
    return lib_min_max(ctx, arguments, token, "max")


def lib_random(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    tr = ctx.transpiler
    min_ = (args and args[0]) or CplInt(token, 0)
    max_ = args[1] if len(args) > 1 else CplInt(token, 2147483647)
    if min_.unique_type.type != "int" or max_.unique_type.type != "int":
        raise_syntax_error("Expected an int argument for Math.random()", token)
    if tr.pack_format >= 18:
        # Just use /random command that got added in 1.20.2
        if isinstance(min_, CplInt) and isinstance(max_, CplInt):
            eid = f"int_{get_uuid()} __temp__"
            ctx.file.append(f"execute store result score {eid} run random value {min_.value}..{max_.value}")
            return CplScore(token, eid)

    if isinstance(min_, CplNBT):
        min_ = min_.cache(ctx, force="score")
    if isinstance(max_, CplNBT):
        max_ = max_.cache(ctx, force="score")
    if "random_setup" not in tr.data:
        tr.files["__math__/random_setup"] = [
            "summon area_effect_cloud ~ ~ ~ {Tags:['__math__.random']}",
            "execute store result score __random__.seed __temp__ run data get entity @e[limit=1,type=area_effect_cloud,tag=__math__.random] UUID[0] 1",
            "kill @e[type=area_effect_cloud,tag=__math__.random]",
            "scoreboard players set __random__.rng.a __temp__ 656891",
            "scoreboard players set __random__.rng.c __temp__ 875773"
        ]
        tr.files["__math__/random"] = [
            "scoreboard players operation __random__.seed __temp__ *= __random__.rng.a __temp__",
            "scoreboard players operation __random__.seed __temp__ += __random__.rng.c __temp__",
            "scoreboard players operation __random__.rng.result __temp__ = __random__.seed __temp__",
            "scoreboard players operation __random__.tmp __temp__ = __random__.rng.result __temp__",
            "scoreboard players operation __random__.rng.result __temp__ %= __random__.rng.bound __temp__",
            "scoreboard players operation __random__.tmp __temp__ -= __random__.rng.result __temp__",
            "scoreboard players operation __random__.tmp __temp__ += __random__.rng.bound __temp__",
            f"execute if score __random__.tmp __temp__ matches ..0 run function {tr.pack_namespace}:__math__/random"
        ]

    if isinstance(min_, CplInt) and isinstance(max_, CplInt):
        ctx.file.append(f"scoreboard players set __random__.rng.bound __temp__ {max_.value - min_.value + 1}")
    elif isinstance(min_, CplScore) and isinstance(max_, CplInt):
        ctx.file.append(f"scoreboard players set __random__.rng.bound __temp__ {max_.value + 1}")
        ctx.file.append(f"scoreboard players operation __random__.rng.bound __temp__ -= {min_.location}")
    elif isinstance(min_, CplInt) and isinstance(max_, CplScore):
        ctx.file.append(f"scoreboard players set __random__.rng.bound __temp__ {-min_.value + 1}")
        ctx.file.append(f"scoreboard players operation __random__.rng.bound __temp__ += {max_.location}")
    else:
        ctx.file.append(f"scoreboard players operation __random__.rng.bound __temp__ = {max_.location}")
        ctx.file.append(f"scoreboard players operation __random__.rng.bound __temp__ -= {min_.location}")
        ctx.file.append(f"scoreboard players add __random__.rng.bound __temp__ 1")
    ctx.file.append(f"function {tr.pack_namespace}:__math__/random")

    return CplScore(token, "__random__.rng.result __temp__", "int")._set_add(ctx, min_)


def pseudo_int(ctx, cpl):
    if isinstance(cpl, CplFloat):
        return CplInt(cpl.token, int(cpl.value * FLOAT_PREC))
    if isinstance(cpl, CplScore):
        return CplScore(cpl.token, cpl.location, "int")
    if isinstance(cpl, CplFloatNBT):
        cached = cpl.cache(ctx, force="score")
        return CplScore(cpl.token, cached.location, "int")
    raise ValueError("Unexpected input")


def lib_frandom(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    if len(args) != 2:
        raise_syntax_error("Expected 2 arguments for Math.frandom()", token)
    if args[0].unique_type.type != "float" or args[1].unique_type.type != "float":
        raise_syntax_error("Expected a float argument for Math.frandom()", token)
    score = lib_random(ctx, list(map(lambda arg: pseudo_int(ctx, arg), args)), token)
    return CplScore(token, score.location, "float")


def lib_ipow(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    tr = ctx.transpiler
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for Math.ipow()", token)
    a0 = args[0]
    a1 = args[1]
    if a0.unique_type.type not in {"int", "float"}:
        raise_syntax_error(f"Expected an int or float for the first argument of Math.ipow()", token)
    if a1.unique_type.type != "int":
        raise_syntax_error(f"Expected an int for the second argument of Math.ipow()", token)
    if isinstance(a0, CplNBT):
        a0 = a0.cache(ctx, force="score")
    if isinstance(a1, CplIntNBT):
        a1 = a1.cache(ctx, force="score")
    if (isinstance(a0, CplInt) or isinstance(a0, CplFloat)) and isinstance(a1, CplInt):
        if isinstance(a0, CplFloat):
            return CplFloat(token, a0.value ** a1.value)
        return CplInt(token, int(a0.value ** a1.value))  # type: ignore
    output = CplScore(token, "__ipow__output __temp__", a0.unique_type.type)

    if isinstance(a1, CplInt):
        if a1.value == 0:
            return CplInt(token, 1)
        if a1.value == 1:
            return a0
        output._set(ctx, a0)
        for i in range(a1.value - 1):
            output._set_mul(ctx, a0)
    else:
        if isinstance(a0, CplInt):
            if a0.value == 0:
                return CplInt(token, 0)
            if a0.value == 1:
                return CplInt(token, 1)
        output._set(ctx, CplInt(token, 1))
        a0.cache(ctx, force="score", score_loc="__ipow__x __temp__")
        a1.cache(ctx, force="score", score_loc="__ipow__n __temp__")
        ctx.file.append(f"function {tr.pack_namespace}:__ipow__/loop{a0.unique_type.type[0]}")
        tr.files[f"__ipow__/loop{a0.unique_type.type[0]}"] = [
            f"scoreboard players remove __ipow__n __temp__ 1",
            f"scoreboard players operation __ipow__output __temp__ *= __ipow__x __temp__",
            f"scoreboard players operation __ipow__output __temp__ /= FLOAT_PREC __temp__" if a0.unique_type.type == "float" else "",
            f"execute if score __ipow__n __temp__ matches 1.. run function {tr.pack_namespace}:__ipow__/loop",
        ]

    return output


def lib_fastexp(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    score = helper_float0(ctx, args, token, "Math.fastexp")
    if isinstance(score, float):
        return CplFloat(token, math.exp(score))

    exp_n = 8
    score._set_div(ctx, CplInt(token, 2 ** exp_n))
    score._set_add(ctx, CplInt(token, 1))
    for i in range(exp_n):
        score._set_mul(ctx, score)

    return score


def lib_exp(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    tr = ctx.transpiler
    score = helper_float0(ctx, args, token, "Math.exp", "__exp__x __temp__")
    if isinstance(score, float):
        return CplFloat(token, math.exp(score))

    ctx.file.append(f"function {tr.pack_namespace}:__exp__/run")

    tr.files["__exp__/init"] = [
        f"scoreboard players set __exp__output __temp__ {int(FLOAT_PREC)}",
        f"scoreboard players set __exp__xn __temp__ {int(FLOAT_PREC)}",
        f"scoreboard players set __exp__nf __temp__ 1",
        f"scoreboard players set __exp__n __temp__ 0",
        f"function {tr.pack_namespace}:__exp__/loop"
    ]

    tr.files["__exp__/loop"] = [
        f"scoreboard players add __exp__n __temp__ 1",
        f"scoreboard players operation __exp__nf __temp__ *= __exp__n __temp__",
        f"scoreboard players operation __exp__xn __temp__ *= __exp__x __temp__",
        f"scoreboard players operation __exp__xn __temp__ /= FLOAT_PREC __temp__",
        f"scoreboard players operation __exp__xmn __temp__ = __exp__xn __temp__",
        f"scoreboard players operation __exp__xmn __temp__ /= __exp__nf __temp__",
        f"scoreboard players operation __exp__output __temp__ += __exp__xmn __temp__",
        f"execute unless score __exp__xmn __temp__ matches 0 run function {tr.pack_namespace}:__exp__/loop",
    ]

    return CplScore(token, "__exp__output __temp__", "float")


def lib_sin_cos(ctx: TranspilerContext, args: List[CompileTimeValue], token, cos):
    name = "cos" if cos else "sin"
    x = helper_float0(ctx, args, token, f"Math.{name}", f"__{name}__x __temp__")
    if isinstance(x, float):
        return CplFloat(token, math.cos(x) if cos else math.sin(x))

    if cos:
        # score = pi/2 - score
        x._set_sub(ctx, CplFloat(token, math.pi / 2))
        x._set_mul(ctx, CplFloat(token, -1))

    # https://en.wikipedia.org/wiki/Bh%C4%81skara_I%27s_sine_approximation_formula

    # sin(x) = 16 * x * (pi - x) / (5pi^2 - 4x(pi - x))
    # x0 = pi
    # x0 -= x
    # x1 = 16
    # x1 *= x
    # x1 *= x0
    # x2 = 5pi^2
    # x3 = 4
    # x3 *= x
    # x3 *= x0
    # x2 -= x3
    # x1 /= x2
    # return x1

    x0 = CplScore(token, f"__{name}__x0 __temp__", "float")
    x1 = CplScore(token, f"__{name}__x1 __temp__", "float")
    x2 = CplScore(token, f"__{name}__x2 __temp__", "float")
    x3 = CplScore(token, f"__{name}__x3 __temp__", "float")

    x0._set(ctx, CplFloat(token, math.pi))
    x0._set_sub(ctx, x)
    x1._set(ctx, CplFloat(token, 16))
    x1._set_mul(ctx, x)
    x1._set_mul(ctx, x0)
    x2._set(ctx, CplFloat(token, 5 * math.pi * math.pi))
    x3._set(ctx, CplFloat(token, 4))
    x3._set_mul(ctx, x)
    x3._set_mul(ctx, x0)
    x2._set_sub(ctx, x3)
    x1._set_div(ctx, x2)

    return x1


def lib_sin(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return lib_sin_cos(ctx, args, token, False)


def lib_cos(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return lib_sin_cos(ctx, args, token, True)


def lib_tan(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.tan")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.tan(x.value))
    return lib_sin(ctx, args, token)._set_div(ctx, lib_cos(ctx, args, token))


def lib_csc(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.csc")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, 1 / math.sin(x.value))
    return CplFloat(token, 1)._div(ctx, lib_sin(ctx, args, token))


def lib_sec(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.sec")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, 1 / math.cos(x.value))
    return CplFloat(token, 1)._div(ctx, lib_cos(ctx, args, token))


def lib_cot(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.cot")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, 1 / math.tan(x.value))
    return CplFloat(token, 1)._div(ctx, lib_tan(ctx, args, token))


def lib_fastarcsin(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    x = helper_float0(ctx, args, token, f"Math.fastarcsin", f"__fastarcsin__x __temp__")
    if isinstance(x, float):
        return CplFloat(token, math.asin(x))
    x2 = CplScore(token, f"__fastarcsin__x2 __temp__", "float")
    num = CplScore(token, f"__fastarcsin__num __temp__", "float")
    denom = CplScore(token, f"__fastarcsin__denom __temp__", "float")

    # x2 = x*x
    # num = ((-0.0187293 * x2 + 0.0742610) * x2 - 0.2121144) * x2 + 1.5707288;
    # denom = ((0.0758669 * x2 + 1.0) * x2 + 1.0);
    # return x * num / denom

    # x2 = x
    # x2 *= x
    # num = -0.0187293
    # num *= x2
    # num += 0.0742610
    # num *= x2
    # num -= 0.2121144
    # num *= x2
    # num += 1.5707288
    # denom = 0.0758669
    # denom *= x2
    # denom += 1.0
    # denom *= x2
    # denom += 1.0
    # x *= num
    # x /= denom
    # return x

    x2._set(ctx, x)
    x2._set_mul(ctx, x)
    num._set(ctx, CplFloat(token, -0.0187293))
    num._set_mul(ctx, x2)
    num._set_add(ctx, CplFloat(token, 0.0742610))
    num._set_mul(ctx, x2)
    num._set_sub(ctx, CplFloat(token, 0.2121144))
    num._set_mul(ctx, x2)
    num._set_add(ctx, CplFloat(token, 1.5707288))
    denom._set(ctx, CplFloat(token, 0.0758669))
    denom._set_mul(ctx, x2)
    denom._set_add(ctx, CplFloat(token, 1.0))
    denom._set_mul(ctx, x2)
    denom._set_add(ctx, CplFloat(token, 1.0))
    x._set_mul(ctx, num)
    x._set_div(ctx, denom)

    return x


def lib_fastarccos(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.fastarccos")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.acos(x.value))
    return (lib_fastarcsin(ctx, args, token)
            ._set_sub(ctx, CplFloat(token, math.pi / 2))
            ._set_mul(ctx, CplFloat(token, -1)))


def lib_fastarctan(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.fastarctan")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.atan(x.value))
    # arctan(x) = arcsin(x / sqrt(1 + x^2))

    x0 = CplScore(token, f"__fastarcsin__x __temp__", "float")

    x0._set(ctx, x)
    x0._set_mul(ctx, x)
    x0._set_add(ctx, CplFloat(token, 1.0))
    x0 = x0._call_index(ctx, "sqrt", [x0], token)
    x._set_div(ctx, x0)

    return lib_fastarcsin(ctx, [x], token)


def lib_fastarccsc(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.fastarccsc")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.asin(1 / x.value))
    return lib_fastarcsin(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


def lib_fastarcsec(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.fastarcsec")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.acos(1 / x.value))
    return lib_fastarccos(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


def lib_fastarccot(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.fastarccot")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.atan(1 / x.value))
    return lib_fastarctan(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


def lib_arcsin(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    x = helper_float0(ctx, args, token, f"Math.arcsin", "__arcsin__x __temp__")
    if isinstance(x, float):
        return CplFloat(token, math.asin(x))

    # d/dx arcsin(x) = 1 / sqrt(1-x^2)
    # arcsin(a) = x
    # a = sin(x)
    # a - sin(x) = 0

    # iterate: y -= (sin(y) - x) / cos(y)

    y = CplScore(token, f"__arcsin__y __temp__", "float")
    z = CplScore(token, f"__arcsin__z __temp__", "float")

    # heuristic guess
    y._set(ctx, x)

    for i in range(2):
        z._set(ctx, lib_sin(ctx, [x], token))
        z._set_sub(ctx, x)
        z._set_div(ctx, lib_cos(ctx, [x], token))
        y._set_sub(ctx, z)

    return y


def lib_arccos(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.arccos")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.acos(x.value))

    return (lib_arcsin(ctx, args, token)
            ._set_sub(ctx, CplFloat(token, math.pi / 2))
            ._set_mul(ctx, CplFloat(token, -1)))


def lib_arctan(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.arctan")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.atan(x.value))

    # arctan(x) = arcsin(x / sqrt(1 + x^2))

    x0 = CplScore(token, f"__arcsin__x __temp__", "float")

    x0._set(ctx, x)
    x0._set_mul(ctx, x)
    x0._set_add(ctx, CplFloat(token, 1.0))
    x0 = x0._call_index(ctx, "sqrt", [x0], token)
    x._set_div(ctx, x0)

    return lib_arcsin(ctx, [x], token)


def lib_arccsc(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.arccsc")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.asin(1 / x.value))

    return lib_arcsin(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


def lib_arcsec(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.arcsec")
    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.acos(1 / x.value))

    return lib_arccos(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


def lib_arccot(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    helper_float0_check(ctx, args, token, f"Math.arccot")

    x = args[0]
    if isinstance(x, CplInt) or isinstance(x, CplFloat):
        return CplFloat(token, math.atan(1 / x.value))

    return lib_arctan(ctx, [CplFloat(token, 1)._div(ctx, x)], token)


factorial_map = [1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880, 3628800, 39916800, 479001600]


def lib_factorial(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    if len(args) != 1:
        raise_syntax_error(f"Expected 1 argument for Math.factorial()", token)
    n = args[0]
    if n.unique_type.type not in {"int", "float"}:
        raise_syntax_error(f"Expected an int or float argument for Math.factorial()", token)
    if isinstance(n, CplInt) or isinstance(n, CplFloat):
        return CplInt(token, factorial_map[int(n.value)])
    if isinstance(n, CplNBT):
        n = n.cache(ctx, force="score", force_t="int")
    ctx.file.append(f"scoreboard players set __factorial__output __temp__ 0")
    for index, value in enumerate(factorial_map):
        ctx.file.append(
            f"execute if score {n.location} matches {index} run scoreboard players set __factorial__output __temp__ {value}")

    return CplScore(token, "__factorial__output __temp__", "int")


add_lib(VariableDeclaration(
    name="Math",
    type=CustomCplObject({
        "sqrt": lib_sqrt,
        "cbrt": lib_cbrt,
        "isqrt": lib_isqrt,
        "floor": lib_floor,
        "ceil": lib_ceil,
        "round": lib_round,
        "min": lib_min,
        "max": lib_max,
        "random": lib_random,
        "frandom": lib_frandom,
        "ipow": lib_ipow,
        "fastexp": lib_fastexp,
        "exp": lib_exp,
        "sin": lib_sin,
        "cos": lib_cos,
        "tan": lib_tan,
        "csc": lib_csc,
        "sec": lib_sec,
        "cot": lib_cot,
        "arcsin": lib_arcsin,
        "arccos": lib_arccos,
        "arctan": lib_arctan,
        "arccsc": lib_arccsc,
        "arcsec": lib_arcsec,
        "arccot": lib_arccot,
        "fastarcsin": lib_fastarcsin,
        "fastarccos": lib_fastarccos,
        "fastarctan": lib_fastarctan,
        "fastarccsc": lib_fastarccsc,
        "fastarcsec": lib_fastarcsec,
        "fastarccot": lib_fastarccot,
        "factorial": lib_factorial,
        "PI": CplFloat(None, math.pi),
        "E": CplFloat(None, math.e),
        "LN2": CplFloat(None, math.log(2)),
        "LN10": CplFloat(None, math.log(10)),
        "LOG2E": CplFloat(None, math.log(math.e, 2)),
        "LOG10E": CplFloat(None, math.log(math.e, 10)),
        "SQRT1_2": CplFloat(None, math.sqrt(0.5)),
        "SQRT2": CplFloat(None, math.sqrt(2)),
    }),
    constant=True
))



from typing import Union, List

from ._base import CompileTimeValue
from .nbt import CplNBT, object_get_index_nbt, val_nbt
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import CplDefObject


class CplObjectNBT(CplNBT):
    def __init__(self, token: Union[Token, None], location: str, type: CplDefObject):
        super().__init__(token, location, type)
        self.unique_type = type

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplInt) or isinstance(index, CplFloat) or isinstance(index, CplString):
            ind = str(index.value)
            if ind not in self.unique_type.content:
                raise_syntax_error(f"Key '{ind}' is not in the object: {self.unique_type}", index.token)
            return val_nbt(self.token, self.location + "." + ind, self.unique_type.content[ind])
        values = self.unique_type.content.values()
        if len(set(map(str, values))) > 1:
            raise_syntax_error(
                "Cannot index into an object with a non-literal value if it has multiple value types. "
                "This is because for the compiler it's impossible to know what the type of the unknown key "
                "will correspond to. Use the myObject[myKey : valueType] syntax instead",
                self.token
            )
        content_type = list(values)[0]

        return object_get_index_nbt(ctx, content_type, self, index)

    def _get_slice(self, ctx, index1, index2, index3, token):
        pass

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        if self.unique_type.class_name is None:
            return None
        cls = self.unique_type.class_name
        return ctx.transpiler.run_function_with_cpl(
            ctx=ctx,
            name=cls + "." + index,
            args=arguments,
            base=self.token,
            class_name=cls
        )

    def _add(self, ctx, cpl):
        if self.unique_type.class_name != cpl.unique_type.class_name:
            return None
        if self.unique_type != cpl.unique_type:
            return None
        # TODO: array merging
        return None

    def _and(self, ctx, cpl):
        return cpl

    def _or(self, ctx, cpl):
        return self


from .string import CplString
from .float import CplFloat
from .int import CplInt



from typing import Union, Any

from ..tokenizer import Token
from ._base import CompileTimeValue


class CplLiteral(CompileTimeValue):
    def __init__(self, token: Union[Token, None], value: Any):
        super().__init__(token)
        self.value = value



from ._base import CompileTimeValue
from .nbt import val_nbt
from .string import CplString
from ..error import raise_syntax_error
from ..nbt_definitions import ENTITIES_OBJ
from ..tokenizer import Token
from ..utils import SELECTOR_TYPE, get_uuid


class CplSelector(CompileTimeValue):
    def __init__(self, token, value):
        # type: (Token | None, str) -> None
        super().__init__(token)
        self.value = value
        self.unique_type = SELECTOR_TYPE

    def get_py_value(self):
        return None

    def _get_index(self, ctx, index):
        if not isinstance(index, CplString):
            return None
        if index.value not in ENTITIES_OBJ.content:
            return None
        return val_nbt(self.token, f"entity {self.value} {index.value}", ENTITIES_OBJ.content[index.value])

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        raise_syntax_error("Selectors cannot be stored", self.token)

    def get_data_str(self, ctx):
        raise_syntax_error("Selectors cannot be used in operations", self.token)

    def tellraw_object(self, ctx):
        return {"selector": self.value}

    def _and(self, ctx, cpl):
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            return self if cpl.value != 0 else CplInt(self.token, 0)
        score_loc = f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players set {score_loc} 0")
        if isinstance(cpl, CplSelector):
            ctx.file.append(f"execute "
                            f"if entity {self.value} "
                            f"if entity {cpl.value} "
                            f"run scoreboard players add {score_loc} 1")
            return CplScore(self.token, score_loc)
        if isinstance(cpl, CplScore):
            ctx.file.append(f"execute "
                            f"unless score {cpl.location} matches 0..0 "
                            f"if entity {self.value} "
                            f"run scoreboard players add {score_loc} 1")
            return CplScore(self.token, score_loc)
        return None

    def _or(self, ctx, cpl):
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            return self if cpl.value == 0 else CplInt(self.token, 1)
        score_loc = f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players set {score_loc} 0")
        if isinstance(cpl, CplSelector):
            ctx.file.append(f"execute if entity {self.value} run scoreboard players add {score_loc} 1")
            ctx.file.append(f"execute if entity {cpl.value} run scoreboard players add {score_loc} 1")
            return CplScore(self.token, score_loc)
        if isinstance(cpl, CplScore):
            ctx.file.append(f"execute if entity {self.value} "
                            f"run scoreboard players add {score_loc} 1")
            ctx.file.append(f"execute unless score {cpl.location} matches 0..0 "
                            f"run scoreboard players add {score_loc} 1")
            return CplScore(self.token, score_loc)
        return None


from .float import CplFloat
from .int import CplInt
from .score import CplScore



from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.nbtarray import CplArrayNBT
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib
from ..utils import CplDefArray, FLOAT_TYPE


def lib_get_pos(ctx: TranspilerContext, args: List[CompileTimeValue], token: Token | None):
    if len(args) != 0:
        raise_syntax_error("Expected 0 arguments", token)
    ctx.transpiler.files["__getpos__/getpos"] = [
        "data modify storage __getpos__ pos set from entity @s Pos",
        "kill @s"
    ]
    ctx.file.append(f"execute summon marker run function {ctx.transpiler.pack_namespace}:__getpos__/getpos")
    return CplArrayNBT(token, "storage __getpos__ pos", CplDefArray(FLOAT_TYPE))


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="getpos",
    function=lib_get_pos
))



import json
import os
import sys
import traceback
from importlib import util
from types import FunctionType
from typing import Any, Dict, List, Union, Literal

from .dp_ast import (
    ENDERS,
    ContinueStatement,
    DefineClassStatement,
    DefineFunctionStatement,
    ExecuteMacroStatement,
    IfFlowStatement,
    ImportStatement,
    InlineStatement,
    IntroduceVariableStatement,
    LoopStatement,
    ReturnStatement,
    ScheduleStatement,
    Statement,
    StatementType,
    COMMANDS,
    chain_tokens,
    make_expr,
    parse_str, parse, DefineEnumStatement
)
from .error import raise_syntax_error, raise_syntax_error_t, show_warning, raise_error, RadonError
from .nbt_definitions import ENTITIES_OBJ
from .tokenizer import (
    BlockIdentifierToken,
    GroupToken,
    SelectorIdentifierToken,
    Token,
    SET_OP,
    INC_OP,
    CMP_OP,
    CMP_COMBINE_OP,
    cpl_def_from_tokens,
    split_tokens,
    tokenize, LambdaFunctionToken
)
from .utils import (
    INT_TYPE,
    CplDef,
    VariableDeclaration,
    TokenType,
    get_uuid,
    INT_LIMIT,
    CplDefArray, reset_expr_id, FLOAT_PREC, CplDefFunction, get_float_limit
)

cwd = os.path.dirname(os.path.abspath(sys.argv[0]))
if sys.argv[0] == "":
    cwd = "/home/pyodide"


def get_lib_contents(path: str):
    with open(cwd + "/" + path, "r") as f:
        return f.read()


def _get_score_type(score):
    if isinstance(score, int):
        return "int"
    elif isinstance(score, float):
        return "float"
    return "float" if score.startswith("float_") else "int"


def transpile_str(code: str):
    (statements, macros) = parse_str(code)

    return Transpiler(statements, macros)


def split_cmp(expr: List[Token]):
    spl = []
    spl_ci = 0
    spl.append([[], None, []])
    for t in expr:
        if t.value in CMP_COMBINE_OP:
            spl_ci = False
            spl.append(t)
            spl.append([[], None, []])
        elif t.value in CMP_OP:
            if spl_ci != 0:
                raise_syntax_error("Unexpected comparison operator", t)
            spl_ci = 2
            sn1 = spl[-1]
            if isinstance(sn1, list):
                sn1[1] = t
        else:
            spl[-1][spl_ci].append(t)
    return spl


def split_fn_args(tokens: List[Token]):
    return list(map(lambda x: chain_tokens(x), split_tokens(tokens, ",")))


class FunctionArgument:
    def __init__(self, name, store, store_via):
        # type: (str, CplScore | CplNBT, str) -> None
        self.name = name
        self.unique_type = store.unique_type.content if store_via in {"stack", "macro"} else store.unique_type
        self.store = store
        self.store_via = store_via  # set, stack, macro


class LoopDeclaration:
    def __init__(self, eid: str, file: List[str], file_path: str, continue_: str):
        self.eid = eid
        self.file = file
        self.file_path = file_path
        self.continue_ = continue_


class FunctionDeclaration:
    def __init__(
            self,
            type: Literal[
                "radon", "python", "python-cpl", "python-raw", "python-cpl-imported", "mcfunction-imported"],
            name: str,
            returns="void",  # type: CplScore | CplNBT | str
            arguments: List[FunctionArgument] = None,
            function: Any = None,
            file_name: str = "",
            class_name: str | None = None,
            raw_args: List[int] = None
    ):
        if arguments is None:
            arguments = []
        if raw_args is None:
            raw_args = []
        self.type = type
        self.name = name
        self.file_name = file_name
        self.returns = returns
        self.arguments: List[FunctionArgument] = arguments
        self.function = function
        self.class_name = class_name
        self.raw_args = raw_args


class ClassDeclaration:
    def __init__(
            self,
            name: str,
            attributes,  # type: CplObject
            methods: List[FunctionDeclaration],
            sample  # type: CplObjectNBT
    ):
        self.id = get_uuid()
        self.name = name
        self.attributes = attributes
        self.methods = methods
        self.sample = sample


class TranspilerContext:
    def __init__(
            self,
            transpiler,  # type: Transpiler
            file_name: str,
            file: List[str],
            radon_path: str,
            function: FunctionDeclaration | None = None,
            loop: LoopDeclaration | Any | None = None,
            class_name: str | None = None
    ):
        self.transpiler = transpiler
        self.file_name = file_name
        self.file = file
        self.radon_path = radon_path
        self.function = function
        self.loop = loop
        self.class_name = class_name

    def clone(self):
        return TranspilerContext(
            self.transpiler,
            self.file_name,
            self.file,
            self.radon_path,
            self.function,
            self.loop,
            self.class_name
        )

    def clone_at(self, file_name: str):
        new = self.clone()
        new.file_name = file_name
        new.file = self.transpiler.files[file_name]
        return new


def _get_def(v: Any | CplDef):
    return v if isinstance(v, CplDef) else v.unique_type


def raw_group_args(token):
    return list(
        map(lambda x: GroupToken(token.code, x[0].start, x[-1].end, x), split_tokens(token.children, ",")))


from .cpl._base import CompileTimeValue
from .cpl.int import CplInt
from .cpl.float import CplFloat
from .cpl.string import CplString
from .cpl.array import CplArray
from .cpl.tuple import CplTuple
from .cpl.object import CplObject
from .cpl.score import CplScore
from .cpl.selector import CplSelector
from .cpl.nbt import CplNBT, val_nbt
from .cpl.nbtobject import CplObjectNBT
from .cpl.nbtarray import CplArrayNBT


class CustomCplObject(CplObject):
    def __init__(self, obj: Dict[str, CompileTimeValue | Any] = None, raw_functions: Dict[str, Any] = None):
        super().__init__(None)
        self.obj = obj or {}
        self.raw_functions = raw_functions or {}

    def _raw_call_index(self, ctx, index, arguments, token):
        if index in self.raw_functions:
            fn = self.raw_functions[index]
            return fn(ctx, arguments, token)
        return None

    def _call_index(self, ctx, index, arguments, token):
        if index in self.obj:
            fn = self.obj[index]
            if isinstance(fn, FunctionType):
                if fn.__code__.co_argcount == 2:
                    return fn(ctx, arguments)
                return fn(ctx, arguments, token)
        return None

    def _get_index(self, ctx, index: CompileTimeValue):
        if (isinstance(index, CplString)
                and index.value in self.obj
                and isinstance(self.obj[index.value], CompileTimeValue)):
            return self.obj[index.value]
        return None


def _type_to_cpl(
        token: Token, type: CplDef, score_loc: str, nbt_loc: str, force_nbt: bool = False
) -> Union[CplScore, CplNBT]:
    if force_nbt or type.type in {"string", "array", "object"}:
        return val_nbt(token, nbt_loc, type)
    return CplScore(token, score_loc, type.type)


def py_to_cpl(v):
    if isinstance(v, CompileTimeValue):
        return v
    if isinstance(v, str):
        return CplString(None, v)
    if isinstance(v, int):
        return CplInt(None, v)
    if isinstance(v, float):
        return CplFloat(None, v)
    if isinstance(v, bool):
        return CplInt(None, 1 if v else 0)
    return CplInt(None, 0)


builtin_fns: Dict[str, List[FunctionDeclaration]] = {}
builtin_vars: Dict[str, VariableDeclaration] = {}


def add_lib(*libs):
    for lib in libs:
        if isinstance(lib, FunctionDeclaration):
            if lib.name in builtin_fns:
                builtin_fns[lib.name].append(lib)
            else:
                builtin_fns[lib.name] = [lib]
        elif isinstance(lib, VariableDeclaration):
            builtin_vars[lib.name] = lib


radon_module_id = 0


def import_module_from_path(file_path):
    global radon_module_id
    radon_module_id += 1
    module_name = f"radon__{radon_module_id}"
    spec = util.spec_from_file_location(module_name, file_path)
    module = util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def get_module_attr(module):
    d = {}
    for k in module.__dict__:
        if k.startswith("__"):
            continue
        v = module.__dict__[k]
        d[k] = v
    return d


# add lib
from .builtin import rmath, print as _no, pyeval, time, swap, stdvar, listener, exit, recipe, success, raycast, getpos, \
    data

_ = [rmath, _no, pyeval, time, swap, stdvar, listener, exit, recipe, success, raycast, getpos, data]


def get_fn_macro_obj(ctx: TranspilerContext):
    if not ctx.function:
        return ""
    d = []
    has_macro = False
    for arg in ctx.function.arguments:
        if arg.store_via == "macro":
            has_macro = True
            d.append(f"'{arg.name}':'$({arg.name})'")
    if not has_macro:
        return ""
    return " {" + ",".join(d) + "}"


class Transpiler:
    def __init__(self, statements: List[Statement], macros: List[Statement], pack_namespace: str = "mypack",
                 pack_description: str = "", pack_format: int = 48, main_dir: str = "./",
                 main_file_path: str = "main.rn", debug_mode=False) -> None:
        reset_expr_id()
        self.files = dict()
        self.dp_files = dict()
        self.pack_namespace = pack_namespace
        self.pack_description = pack_description
        self.pack_format = pack_format
        self.main_dir = main_dir
        self.variables: Dict[str, VariableDeclaration] = dict()
        self.functions: List[FunctionDeclaration] = []
        self.classes: Dict[str, ClassDeclaration] = dict()
        self.loops: Dict[Any, LoopDeclaration] = dict()
        self.init_libs = []
        self.data = dict()
        self.main_file = []
        self.tick_file = []
        self.tempFiles: Dict[str, str] = dict()
        self.builtin_fns = builtin_fns
        self.builtin_vars = builtin_vars
        self.s = "" if self.pack_format >= 48 else "s"
        self.debug_mode = debug_mode

        if len(statements) == 0:
            self.files = {}
            return
        self.macros = macros
        load_file = [
            'scoreboard objectives add __temp__ dummy',
            f'scoreboard players set FLOAT_PREC __temp__ {FLOAT_PREC}'
        ]
        self.files["__load__"] = load_file
        self.load_file = load_file
        main_file = []
        self.main_file = main_file
        self._transpile(
            TranspilerContext(
                transpiler=self,
                file_name="__main__",
                file=main_file,
                radon_path=main_file_path,
                function=None,
                loop=None),
            statements
        )
        load_file.extend(main_file)

        if "tick" not in self.files:
            if self.tick_file:
                self.files["tick"] = self.tick_file
        else:
            self.files["tick"] = self.tick_file + self.files["tick"]

        if len(self.variables.keys()) > 0:
            self.load_file.insert(0, 'scoreboard objectives add global dummy "global"')

        # remove the lines after immediate returns
        for file in self.files:
            lines = self.files[file]
            if len(lines) == 0:
                continue
            index = 0
            for index, line in enumerate(lines):
                if line.startswith("return "):
                    break
            self.files[file] = lines[:index + 1]
        if self.debug_mode:
            for file in self.files:
                lines = self.files[file]
                new_file = []
                for line in lines:
                    new_file.append(line)
                    if line.startswith("scoreboard players "):
                        spl = line.split(" ")
                        new_file.append('tellraw @a ["' + spl[3] + " " + spl[4] + ' = ",{"score":{"name":"' + spl[
                            3] + '","objective":"' + spl[4] + '"}}]')
                self.files[file] = new_file

    def get_datapack_files(self):
        fn_folder = "function" + self.s
        dp_files = dict()
        for file in self.files:
            dp_files[f"data/{self.pack_namespace}/{fn_folder}/{file}.mcfunction"] = "\n".join(self.files[file])
        dp_files["pack.mcmeta"] = json.dumps({
            "pack": {
                "pack_format": self.pack_format,
                "description": self.pack_description,
            }
        }, indent=2)
        dp_files[f"data/minecraft/tags/{fn_folder}/load.json"] = json.dumps(
            {
                "values": [f"{self.pack_namespace}:__load__"]
            }, indent=2)
        if "tick" in self.files:
            dp_files[f"data/minecraft/tags/{fn_folder}/tick.json"] = json.dumps(
                {
                    "values": [f"{self.pack_namespace}:tick"]
                }, indent=2)
        for f in self.dp_files:
            dp_files[f"data/{self.pack_namespace}/{f}"] = self.dp_files[f]
        return dp_files

    def get_literal_num(self, num):
        isf = isinstance(num, float)
        if isf:
            num = int(num * FLOAT_PREC)
        k = f"literal_num_{num}"
        if k not in self.data:
            self.data[k] = True
            self.load_file.append(f"scoreboard players set literal_{num} __temp__ {num}")
        return CplScore(None, f"literal_{num}", "float" if isf else "int")

    def get_temp_file_name(self, content: str | List[str]):
        if isinstance(content, list):
            content = "\n".join(content)
        if content in self.tempFiles:
            return self.tempFiles[content]
        file_name = f"__temp__/{get_uuid()}"
        self.tempFiles[content] = file_name
        self.files[file_name] = content.split("\n")
        return file_name

    def get_all_fn_by_name(self, name):
        ls = []
        for fn in self.functions:
            if fn.name == name:
                ls.append(fn)
        return ls

    def check_args(self, args1, args2, base):
        if len(args1) != len(args2):
            return False
        for index, arg1 in enumerate(args1):
            arg2 = args2[index]
            if isinstance(_get_def(arg2), CplDefFunction) and isinstance(arg1, CplString) and arg1.is_fn_reference:
                temp = arg1
                arg1 = arg2
                arg2 = temp
            t1 = _get_def(arg1)
            t2 = _get_def(arg2)
            if isinstance(t1, CplDefFunction) and isinstance(arg2, CplString) and arg2.is_fn_reference:
                arg_fn = self.get_fn(arg2.value, t1.arguments, base)
                if not arg_fn:
                    return False
                if arg_fn.returns == "auto":
                    raise_syntax_error("Cannot use function as a value before the return type is not determined", base)
                if arg_fn.returns == "void" and t1.returns is None:
                    continue
                if arg_fn.returns.unique_type != t1.returns:
                    continue
                continue
            if t1 != t2:
                return False
        return True

    def assert_args(self, name, args1, args2, base):
        if not self.check_args(args1, args2, base):
            raise_syntax_error(
                f"Invalid arguments. Expected usage: {name}({', '.join(map(str, args1))}), but got: {name}({', '.join(map(str, args2))})",
                base)

    def get_fn(self, name, arguments: List[CompileTimeValue | CplDef], base) -> FunctionDeclaration | None:
        for fn in self.functions:
            if fn.name == name and self.check_args(fn.arguments, arguments, base):
                return fn
        return None

    def fn_exists(self, name):
        for fn in self.functions:
            if fn.name == name:
                return fn
        return None

    def _transpile_statement(self, ctx: TranspilerContext, statement: Statement):
        if isinstance(statement, DefineEnumStatement):
            values = {}
            for k in statement.values:
                v = statement.values[k]
                if isinstance(v, int):
                    values[k] = CplInt(None, v)
                elif isinstance(v, float):
                    values[k] = CplFloat(None, v)
                else:
                    values[k] = CplString(None, v)
            self.variables[statement.name.value] = VariableDeclaration(
                statement.name.value, CplObject(None, values), True)
            return
        if isinstance(statement, DefineClassStatement):
            if ctx.function or ctx.loop:
                raise_syntax_error(
                    "Classes should be declared in the main scope", statement
                )
            name = statement.name.value
            if name in self.classes or self.fn_exists(name):
                raise_syntax_error(f"Class {name} is already defined", statement)
            base_object = CplObject(statement.name, dict(), name)
            for (attr_name, attr_expr) in statement.attributes:
                cpl = self.tokens_to_cpl(ctx, attr_expr)
                base_object.value[attr_name] = cpl
                base_object.unique_type.content[attr_name] = cpl.unique_type
            sample: CplObjectNBT = base_object.cache(ctx, nbt_loc=f"storage class_sample {name}",  # type: ignore
                                                     force="nbt")

            class_ctx = TranspilerContext(
                transpiler=self,
                file_name=ctx.file_name,
                file=ctx.file,
                radon_path=ctx.radon_path,
                function=None,
                loop=None,
                class_name=name
            )

            methods = []
            has_init = False
            cls = ClassDeclaration(name, base_object, methods, sample)
            self.classes[name] = cls
            for mt in statement.methods:
                if mt.name.value == name:
                    has_init = True
                args = self._chains_to_args(mt.arguments)
                fn_dec = FunctionDeclaration(
                    type="radon",
                    name=name + "." + mt.name.value,
                    returns="auto",
                    arguments=args,
                    file_name=statement.name.value,
                    function="replace me"
                )
                self.functions.append(fn_dec)
                methods.append(fn_dec)

            new_methods = []

            for mt in statement.methods:
                self._transpile_statement(class_ctx, mt)
                fn = self.functions[-1]
                new_methods.append(fn.name)
            cls.methods = new_methods

            if not has_init:
                self._transpile(class_ctx, parse(
                    tokenize("fn " + name + "() {}")[0],
                    [], list(self.classes.keys())))
                fn = self.functions[-1]
                methods.append(fn)
            return
        if isinstance(statement, ImportStatement):
            pt = statement.path.value[1:-1]
            if pt.endswith(".py"):
                bef = os.getcwd()
                os.chdir(os.path.dirname(ctx.radon_path))
                if not os.path.exists(pt) or not os.path.isfile(pt):
                    os.chdir(bef)
                    raise_syntax_error("File not found", statement)
                try:
                    lib_module = import_module_from_path(pt)
                except Exception as e:
                    os.chdir(bef)
                    raise_error("Import error", str(e), statement)
                    raise e
                attrs = get_module_attr(lib_module)
                os.chdir(bef)
                if statement.as_ is not None:
                    name = statement.as_.value
                    if name not in self.variables:
                        self.variables[name] = VariableDeclaration(name, CustomCplObject(attrs), True)
                    return True
                for name in attrs:
                    v: Any = attrs[name]
                    if isinstance(v, FunctionType) and v.__code__.co_argcount in {2, 3}:
                        self.functions.append(FunctionDeclaration(
                            type="python-cpl-imported",
                            name=name,
                            function=v
                        ))
                    elif isinstance(v, CompileTimeValue):
                        if name not in self.variables:
                            self.variables[name] = VariableDeclaration(name, v, True)
                return True
            if pt.endswith(".mcfunction"):
                if statement.as_ is None:
                    raise_syntax_error("Expected 'as' in mcfunction import statement", statement)
                bef = os.getcwd()
                os.chdir(os.path.dirname(ctx.radon_path))
                if not os.path.exists(pt) or not os.path.isfile(pt):
                    os.chdir(bef)
                    raise_syntax_error("File not found", statement)
                content = open(pt, "r", encoding="utf-8").read()
                os.chdir(bef)
                fn_id = get_uuid()
                self.files[f"__imported__/{fn_id}"] = content.split("\n")
                self.functions.append(FunctionDeclaration(
                    type="mcfunction-imported",
                    name=statement.as_.value,
                    file_name=f"__imported__/{fn_id}"
                ))
                return True
            if pt.endswith(".rn"):
                if statement.as_ is not None:
                    raise_syntax_error("Unexpected 'as' in radon file import statement", statement)
                bef = os.getcwd()
                os.chdir(os.path.dirname(ctx.radon_path))
                if not os.path.exists(pt) or not os.path.isfile(pt):
                    os.chdir(bef)
                    raise_syntax_error("File not found", statement)
                content = open(pt, "r", encoding="utf-8").read()
                os.chdir(bef)
                (tokens, macros) = tokenize(content)
                ctx.transpiler._transpile(
                    TranspilerContext(
                        transpiler=ctx.transpiler,
                        file_name="__main__",
                        file=self.main_file,
                        radon_path=pt,
                        function=None,
                        loop=None,
                        class_name=None
                    ), parse(tokens, macros, list(ctx.transpiler.classes.keys()))
                )
                return True
            raise_syntax_error("Invalid import", statement)
        if isinstance(statement, ScheduleStatement):
            file_name = self._run_safe("__schedule__", statement.body, ctx)
            ctx.file.append(
                f"schedule function {self.pack_namespace}:{file_name} {statement.time.value} replace"
            )
            _return_safe(ctx)
            return True
        if statement.type == StatementType.BREAK:
            if not ctx.loop:
                raise_syntax_error("Cannot use break outside of loop", statement)
            if ctx.file is ctx.loop.file:
                ctx.file.append("return 0")
            else:
                loop_id = ctx.loop.eid
                ctx.file.append(f"scoreboard players set __break__{loop_id} __temp__ 1")
                ctx.file.append(f"return 0")
            return True
        if isinstance(statement, ContinueStatement):
            if not ctx.loop:
                raise_syntax_error("Cannot use continue outside of loop", statement)
            if ctx.file is ctx.loop.file:
                ctx.file.append(ctx.loop.continue_)
                ctx.file.append("return 0")
            else:
                loop_id = ctx.loop.eid
                ctx.file.append(f"scoreboard players set __continue__{loop_id} __temp__ 1")
                ctx.file.append(f"return 0")
            return True
        if isinstance(statement, LoopStatement):
            file_name = self._run_safe(
                "__loop__",
                statement.body,
                TranspilerContext(
                    transpiler=self,
                    file_name=ctx.file_name,
                    file=ctx.file,
                    radon_path=ctx.radon_path,
                    function=ctx.function,
                    loop=statement)
            )
            mac = get_fn_macro_obj(ctx)
            ctx.file.append(f"{'$' if mac else ''}function {self.pack_namespace}:{file_name}" + mac)
            loop_file = self.files[file_name]
            loop_id = int(file_name.split("/")[-1])
            loop_file.insert(0, f"scoreboard players set __break__{loop_id} __temp__ 0")
            loop_file.insert(0, f"scoreboard players set __continue__{loop_id} __temp__ 0")
            loop_c = self.loops[loop_id]
            loop_file.append(loop_c.continue_)
            return True
        if isinstance(statement, IfFlowStatement):
            has_if = len(statement.body) != 0
            has_else = statement.elseBody and len(statement.elseBody) != 0
            if not has_if and not has_else:
                return True
            resp = self.tokens_to_cpl(ctx, statement.condition)
            if (
                    isinstance(resp, CplInt)
                    or isinstance(resp, CplFloat)
                    or (resp.unique_type.type not in {"int", "float"} and not isinstance(resp, CplSelector))):
                is_true = (not isinstance(resp, CplInt) and not isinstance(resp, CplFloat)) or resp.value != 0
                if is_true:
                    if not has_if:
                        return True
                    self._transpile(ctx, statement.body)
                elif (
                        has_else and statement.elseBody
                ):  # I know this is bad, but the linter wants it (statement.elseBody)
                    self._transpile(ctx, statement.elseBody)
                return True
            if isinstance(resp, CplNBT):
                resp = resp.cache(ctx, force="score")
            if isinstance(resp, CplScore):
                if_cmd = f"execute unless score {resp.location} matches 0..0 run "
                unless_cmd = f"execute if score {resp.location} matches 0..0 run "
            elif isinstance(resp, CplSelector):
                if_cmd = f"execute if entity {resp.value} run "
                unless_cmd = f"execute unless entity {resp.value} run "
            else:
                raise "Invalid condition"
            has_one = (has_if or has_else) and (not has_if or not has_else)
            if has_one:
                one_body = statement.body if has_if else statement.elseBody
                i_cmd = if_cmd if has_if else unless_cmd
                if (
                        ctx.loop
                        and one_body
                        and ctx.loop.file is ctx.file
                        and len(one_body) == 1
                        and one_body[0].type == StatementType.BREAK
                ):
                    ctx.file.append(f"{i_cmd}return 0")
                    return True

            if has_if:
                if_name = self._run_safe("__if__", statement.body, ctx)
                ctx.file.append(f"{if_cmd}function {self.pack_namespace}:{if_name}")
            if has_else:
                else_name = self._run_safe("__else__", statement.elseBody, ctx)
                ctx.file.append(f"{unless_cmd}function {self.pack_namespace}:{else_name}")
            _return_safe(ctx)
            return True
        if isinstance(statement, InlineStatement):
            expr_tokens = statement.exprTokens
            ret = self.tokens_to_cpl(ctx, expr_tokens)
            if isinstance(ret, CplScore):
                # Cleaning the return results of the inline expression since they aren't going to be used
                cl1 = f"scoreboard players operation {ret.location} = "
                cl2 = f"execute store result score {ret.location} run "
                cl3 = f"execute store result score {ret.location} run data get"
                cl4 = f"execute store result score {ret.location} run scoreboard players get"
                if len(ctx.file) > 0 and (
                        ctx.file[-1].startswith(cl1)
                        or ctx.file[-1].startswith(cl3)
                        or ctx.file[-1].startswith(cl4)
                ):
                    ctx.file.pop()
                    return True
                if len(ctx.file) > 0 and ctx.file[-1].startswith(cl2):
                    ctx.file[-1] = ctx.file[-1][len(cl2):]
                if len(ctx.file) > 1 and ret.location not in ctx.file[-1] and ctx.file[-2].startswith(cl1):
                    ctx.file.pop(-2)
            if isinstance(ret, CplNBT) and ret.location.startswith("storage temp "):
                # Cleaning the return results of the inline expression since they aren't going to be used
                cl1 = f"data modify {ret.location} set from"
                if len(ctx.file) > 0 and ctx.file[-1].startswith(cl1):
                    ctx.file.pop()
                    return True
                if len(ctx.file) > 1 and ctx.file[-2].startswith(cl1) and ret.location not in ctx.file[-1]:
                    ctx.file.pop(-2)
                    return True
            return True
        if isinstance(statement, DefineFunctionStatement):
            fn_name = statement.name.value
            if ctx.class_name is not None and fn_name != ctx.class_name:
                fn_name = f"{ctx.class_name}.{fn_name}"
            arguments: List[FunctionArgument] = self._chains_to_args(statement.arguments)
            ret = (
                cpl_def_from_tokens(self.classes, statement.returns)
                if not isinstance(statement.returns, str)
                else statement.returns
            )
            if ret is None:
                raise_syntax_error("Invalid return type for a function", statement)
                assert False
            is_class_init = fn_name == ctx.class_name
            ret_loc = ret
            if isinstance(ret_loc, CplDef):
                assert isinstance(statement.returns, list)
                assert isinstance(ret, CplDef)
                ret_loc = _type_to_cpl(
                    statement.returns[0],
                    ret,
                    score_loc=f"fn_return __temp__",
                    nbt_loc=f"storage temp fn_return"
                )
            if is_class_init:
                if isinstance(statement.returns, list) and len(statement.returns) > 0:
                    raise_syntax_error("Class initializers cannot have return types", statement.returns[0])
                cls = self.classes[fn_name]
                ret_loc = CplObjectNBT(statement.name, f"storage variables this[-1]",
                                       cls.attributes.unique_type)
            count = 0
            new_functions = []
            for f in self.functions:
                if fn_name == f.name:
                    count += 1
                    if self.check_args(f.arguments, arguments, statement.name):
                        if f.function == "replace me":
                            count -= 1
                            continue
                        raise_syntax_error(
                            "Function with the same name and arguments already exists",
                            statement.name
                        )
                new_functions.append(f)
            self.functions.clear()
            self.functions.extend(new_functions)
            file_name = statement.name.value.lower()
            if ctx.class_name is not None:
                cls = self.classes[ctx.class_name]
                file_name = "__class__/" + cls.name.lower() + "_" + file_name
            base_file_name = file_name
            file_name_counter = 0
            while file_name in self.files:
                file_name_counter += 1
                file_name = base_file_name + str(file_name_counter)
            f = FunctionDeclaration(
                type="radon",
                name=fn_name,
                returns=ret_loc,
                arguments=arguments,
                file_name=file_name,
                class_name=ctx.class_name
            )

            self.functions.append(f)
            fn_file = []
            self.files[file_name] = fn_file
            for index, arg in enumerate(arguments):
                if arg.store_via == "macro":
                    t = arg.store.unique_type.content  # type: CplDef
                    macro = f"$({arg.store.location.split(' ')[-1]})"
                    if t.type == "string":
                        macro = f'"{macro}"'
                    fn_file.append(f"$data modify {arg.store.location} append value {macro}")
            self._transpile(
                TranspilerContext(
                    transpiler=self,
                    file_name=file_name,
                    file=fn_file,
                    radon_path=ctx.radon_path,
                    function=f,
                    loop=ctx.loop,
                    class_name=ctx.class_name),
                statement.body
            )
            if is_class_init:
                cls = self.classes[ctx.class_name]
                fn_file.insert(0, f"data modify storage variables this append {cls.sample.get_data_str(ctx)}")
            if f.returns == "auto":
                f.returns = "void"
            fn_file.insert(0, f"scoreboard players set __returned__ __temp__ 0")
            return True
        if isinstance(statement, ReturnStatement):
            if not ctx.function:
                raise_syntax_error(
                    "Cannot use return keyword outside functions", statement
                )
                return True
            expr_tokens = statement.exprTokens
            if ctx.function.returns == "void" and len(expr_tokens) != 0:
                raise_syntax_error(
                    f"Function cannot return a value because it has a void return type",
                    statement
                )
            if ctx.function.returns != "void":
                cpl = self.tokens_to_cpl(ctx, expr_tokens)
                if ctx.function.returns == "auto":
                    ctx.function.returns = cpl.cache(ctx)
                else:
                    if cpl.unique_type != ctx.function.returns.unique_type:
                        raise_syntax_error(
                            f"Function has a return type of {ctx.function.returns.unique_type}, "
                            f"but a {cpl.unique_type} was returned",
                            statement
                        )
                    ctx.function.returns._set(ctx, cpl)

            if ctx.file is not self.files[ctx.function.file_name]:
                ctx.file.append(f"scoreboard players set __returned__ __temp__ 1")
            ctx.file.append(f"return 0")
            return True
        if isinstance(statement, IntroduceVariableStatement):
            name = statement.name.value
            if name in self.variables:
                raise_syntax_error("Variable is already introduced", statement)
            type_v = cpl_def_from_tokens(self.classes, statement.varType)
            if not type_v:
                raise_syntax_error(f"Invalid type", statement)
                assert False
            self.variables[name] = VariableDeclaration(name, type_v, False)
            self.load_file.append(
                f'scoreboard objectives add {name} dummy "{name}"'
            )
            self.load_file.append(
                f"scoreboard players enable {name} global"
            )
            return True
        if isinstance(statement, ExecuteMacroStatement):
            exec_name = self._run_safe("__execute__", statement.body, ctx)
            cmd_str, has_repl = self.proc_cmd(ctx, statement.command, statement)
            mac = get_fn_macro_obj(ctx)
            if has_repl:
                fid = f"__execute__/{get_uuid()}"
                self.files[fid] = [f"$execute {cmd_str} run function {self.pack_namespace}:{exec_name}{mac}"]
                if mac:
                    for arg in ctx.function.arguments:
                        if arg.store_via == "macro":
                            ctx.file.append(f"$data modify storage cmd_mem {arg.name} set value '$({arg.name})'")
                ctx.file.append(f"function {self.pack_namespace}:{fid} with storage cmd_mem")
            else:
                ctx.file.append(
                    f"{'$' if mac else ''}execute {cmd_str} run function {self.pack_namespace}:{exec_name}{mac}")
            _return_safe(ctx)
            return True
        raise_syntax_error("Invalid statement", statement)

    def _transpile(self, ctx: TranspilerContext, statements: List[Statement]):
        for statement in statements:
            self._transpile_statement(ctx, statement)

    def _run_safe(self, folder_name, statements, ctx: TranspilerContext):
        eid = get_uuid()
        file_name = f"{folder_name}/{eid}"
        new_file = []
        loop = ctx.loop
        if isinstance(loop, LoopStatement):
            st = loop
            mac = get_fn_macro_obj(ctx)
            continue_ = f"{'$' if mac else ''}function {self.pack_namespace}:{file_name}" + mac
            if st.time:
                if mac:
                    raise_syntax_error(
                        "Cannot use timed loops inside macro functions, move the loop into a normal function and run it there.",
                        st.time)
                continue_ = f"schedule function {self.pack_namespace}:{file_name} {st.time.value} replace"
            loop = LoopDeclaration(
                eid=eid,
                file=new_file,
                file_path=file_name,
                continue_=continue_
            )
            if st.step:
                step_name = f"__loop__/__step__/{get_uuid()}"
                loop.continue_ = f"{'$' if mac else ''}function {self.pack_namespace}:{step_name}{mac}"
                self.files[step_name] = []
                self._transpile(TranspilerContext(
                    transpiler=self,
                    file_name=step_name,
                    file=self.files[step_name],
                    radon_path=ctx.radon_path,
                    function=ctx.function,
                    loop=loop), st.step)
                self.files[step_name].append(continue_)
            self.loops[eid] = loop
        self.files[file_name] = new_file
        self._transpile(
            TranspilerContext(
                transpiler=self,
                file_name=file_name,
                file=new_file,
                radon_path=ctx.radon_path,
                function=ctx.function,
                loop=loop), statements
        )
        return file_name

    def _chains_to_args(self, chains: List[List[List[Token]]]):
        arg_names = []
        arguments = []
        for arg in chains:
            store_via = "stack"
            if len(arg) > 0 and arg[0][0].value == "$":
                store_via = "macro"
                arg = arg[1:]
            if len(arg) != 2 or arg[1][0].type != TokenType.IDENTIFIER:
                raise_syntax_error("Invalid argument for a function", arg[0][0])
            arg_type_parsed = cpl_def_from_tokens(self.classes, arg[0])
            if arg_type_parsed is None:
                raise_syntax_error("Invalid argument type for a function", arg[0][0])
                assert False
            arg_name = arg[1][0].value
            if arg_name in arg_names:
                raise_syntax_error("Argument name is already used", arg[1][0])
            arguments.append(
                FunctionArgument(
                    arg_name,
                    _type_to_cpl(
                        arg[0][0],
                        CplDefArray(arg_type_parsed) if store_via in {"stack", "macro"} else arg_type_parsed,
                        score_loc="",
                        nbt_loc=f"storage fn_args {arg_name}",
                        force_nbt=True
                    ),
                    store_via=store_via
                )
            )
        return arguments

    def proc_cmd(self, ctx: TranspilerContext, cmd: str, base):
        cmd_str = ""
        i = 0
        has_repl = False
        repl_i = 0
        cmd = " ".join(s[:-1] if s[-1] == "\\" else s for s in cmd.strip().split("\n"))
        while i < len(cmd):
            if cmd[i: i + 3] == "\\$(":
                if self.pack_format < 18:
                    raise_syntax_error_t(
                        f"Macros are not supported in the pack format {self.pack_format}. "
                        f"Consider using a version > (17 or 1.20.1 or 23w35a)",
                        cmd,
                        i,
                        i + 3,
                    )
                cmd_str += "$("
                i += 3
                continue
            if (
                    cmd[i: i + 2] == "$("
                    or cmd[i: i + 5] == "$str("
                    or cmd[i: i + 6] == "$jstr("
                    or cmd[i: i + 6] == "$dstr("
                    or cmd[i: i + 5] == "$dat("
                    or cmd[i: i + 5] == "$loc("
            ):
                si = i
                k = 0
                fp = 0
                while i < len(cmd):
                    i += 1
                    c = cmd[i]
                    if c == "(":
                        if fp == 0:
                            fp = i
                        k += 1
                    if c == ")":
                        k -= 1
                        if k == 0:
                            break
                if cmd[i] != ")":
                    raise_syntax_error_t("Unterminated parentheses", cmd, si, i + 1)
                repl = cmd[fp + 1: i]
                (expr_tokens, _) = tokenize(repl)
                expr_tokens = expr_tokens[:-1]
                try:
                    val = self.tokens_to_cpl(ctx, expr_tokens)
                except RadonError as e:
                    raise_syntax_error(e.text,
                                       Token(base.code, TokenType.POINTER, base.start + si, base.start + i))
                    raise e

                if cmd[si + 1] == "(":
                    cmd_str += f"$(_{repl_i})"
                    val.cache(
                        ctx, nbt_loc=f"storage cmd_mem _{repl_i}", force="nbt"
                    )

                    has_repl = True
                    repl_i += 1

                elif cmd[si + 1] == "s":
                    cmd_str += val.tellraw_object_str(ctx)
                elif cmd[si + 1:si + 1] == "ds":
                    py_val = val.get_py_value()
                    if py_val is not None:
                        cmd_str += str(py_val)
                    else:
                        cmd_str += val.tellraw_object_str(ctx)
                elif cmd[si + 1] == "j":
                    cmd_str += val.tellraw_object_str(ctx)
                elif cmd[si + 1:si + 2] == "da":
                    cmd_str += val.get_data_str(ctx)
                elif cmd[si + 1] == "l":
                    if not isinstance(val, CplScore) and not isinstance(val, CplNBT):
                        val = val.cache(ctx, force="nbt")
                    cmd_str += val.location

                i += 1
                continue
            cmd_str += cmd[i]
            i += 1

        return cmd_str, has_repl

    def run_cmd(self, ctx: TranspilerContext, pointer: Token, score_loc=None, type="result") -> CompileTimeValue:
        file = ctx.file
        cmd = pointer.value
        cmd_str, has_repl = self.proc_cmd(ctx, cmd, pointer)

        eid = score_loc or "int_" + str(get_uuid()) + " __temp__"
        eid_val = CplScore(pointer, eid, "int")

        if not has_repl:
            file.append(f"execute store {type} score {eid} run {cmd_str}")
            return eid_val

        cmd_file = []
        file_name = None
        '''for i in self.files:
            if "\n".join(self.files[i]) == "\n".join(cmd_file) and i.startswith("__cmd__/"):
                file_name = i
                break'''
        if file_name is None:
            cmd_id = get_uuid()
            file_name = f"__cmd__/{cmd_id}"
        self.files[file_name] = cmd_file
        cmd_file.append("$return run " + cmd_str)
        file.append(
            f"execute store {type} score {eid} run function {self.pack_namespace}:{file_name} with storage cmd_mem"
        )
        return eid_val

    def _nbt_var_loc(self, ctx: TranspilerContext, token: Token) -> str:
        if isinstance(token, SelectorIdentifierToken):
            return f"entity {token.selector.value} {token.name.value}"
        if isinstance(token, BlockIdentifierToken):
            spl = split_tokens(token.block.children, ",")
            if len(spl) != 3:
                raise_syntax_error("Invalid block identifier", token)
                raise ValueError("")
            pos: List[str] = []
            for s in spl:
                sym = ""
                if s[0].value in {"~", "^"}:
                    sym = s[0].value
                    s = s[1:]
                cpl = self.tokens_to_cpl(ctx, s)
                if not isinstance(cpl, CplInt) and not isinstance(cpl, CplFloat):
                    raise_syntax_error(
                        f"Block identifier's position has to have 3 literal numbers", token
                    )
                    assert False
                pos.append(sym + str(cpl.value))
            return f"block {' '.join(pos)} {token.name.value}"
        raise ValueError("Invalid token")

    def _var_to_cpl_or_none(
            self, ctx: TranspilerContext, t: Token
    ) -> Union[CompileTimeValue, None]:
        if isinstance(t, SelectorIdentifierToken):
            name = t.name.value
            if name in ENTITIES_OBJ.content:
                return _type_to_cpl(
                    t,
                    ENTITIES_OBJ.content[name],
                    score_loc="",
                    nbt_loc=self._nbt_var_loc(ctx, t),
                    force_nbt=True
                )
            if name not in self.variables:
                return None
            var = self.variables[name]
            var_type = var.type
            return _type_to_cpl(
                t,
                var_type,
                score_loc=f"{t.selector.value} {name}",
                nbt_loc=self._nbt_var_loc(ctx, t),
                force_nbt=False
            )
        if isinstance(t, BlockIdentifierToken):
            name = t.name.value
            if name not in self.variables:
                return None
            var = self.variables[name]
            var_type = var.type
            val = _type_to_cpl(
                t,
                var_type,
                score_loc="",
                nbt_loc=self._nbt_var_loc(ctx, t),
                force_nbt=True
            )
            return val
        if t.type == TokenType.IDENTIFIER:
            name = t.value
            if name not in self.variables and name not in builtin_vars:
                if self.fn_exists(name):
                    return CplString(t, name, is_fn_reference=True)
                if ctx.function and ctx.class_name is not None and name == "this":
                    return CplObjectNBT(t, f"storage variables this[-1]",
                                        self.classes[ctx.class_name].sample.unique_type)
                if ctx.function:
                    fn = ctx.function
                    found = None
                    for arg in fn.arguments:
                        if arg.name == name:
                            found = arg
                            break
                    if found:
                        if found.store_via in {"stack", "macro"}:
                            return val_nbt(found.store.token, found.store.location + "[-1]",
                                           found.store.unique_type.content)
                        return found.store
                return None
            var = self.variables[name] if name in self.variables else builtin_vars[name]
            if var.init_func is not None:
                var.init_func(ctx)
            if var.get_func is not None:
                return var.get_func(ctx)
            var_type = var.type
            if var.constant:
                return var.value
            return _type_to_cpl(
                t,
                var_type,
                score_loc=f"{name} global",
                nbt_loc=f"storage variables {name}",
            )
        assert False

    def variable_token_to_cpl(
            self, ctx: TranspilerContext, t: Token
    ) -> CompileTimeValue:
        res = self._var_to_cpl_or_none(ctx, t)
        if res is None:
            raise_syntax_error(f"Variable '{t.value}' was not found", t)
            assert False
        return res

    def chain_to_cpl(self, ctx: TranspilerContext, t: List[Token]) -> CompileTimeValue:
        cpl = self.token_to_cpl(ctx, t[0])
        for token in t[1:]:
            if token.type == TokenType.IDENTIFIER:
                cpl = cpl.get_index(ctx, CplString(token, token.value), token)
                continue
            if token.type == TokenType.INT_LITERAL:
                cpl = cpl.get_index(ctx, CplInt(token, token.value), token)
                continue
            if isinstance(token, GroupToken):
                if token.open.value == "[":
                    if len(token.children) == 0:
                        raise_syntax_error("A bracket index has to include an expression inside.", token)
                        assert False
                    if len(token.children) == 1:
                        t0 = token.children[0]
                        if (isinstance(t0, GroupToken)
                                and t0.open.value == "{"
                                and isinstance(cpl, CplArrayNBT)):
                            cpl = val_nbt(t0, cpl.location + token.value, cpl.unique_type.content)
                            continue

                    spl = list(map(lambda x: self.tokens_to_cpl(ctx, x) if len(x) > 0 else None,
                                   split_tokens(token.children, ":", comma_errors=False)))
                    if len(spl) == 1:
                        cpl = cpl.get_index(ctx, spl[0], token)
                        continue
                    if len(spl) > 3:
                        raise_syntax_error("A slice can only include two colons.", token)
                        assert False
                    while len(spl) < 3:
                        spl.append(None)
                    if spl[0] is None:
                        spl[0] = CplInt(token, 0)
                    if spl[1] is None:
                        spl[1] = CplInt(token, -1)
                    if spl[2] is None:
                        spl[2] = CplInt(token, 1)

                    cpl = cpl.get_slice(ctx, spl[0], spl[1], spl[2], token)
                    continue
                if token.func:
                    cpl = cpl.call_index(ctx, token.func.value, token, token)
                    continue
                if token.open.value == "(":
                    cpl = cpl.call(ctx, token, token)
                    continue
            raise_syntax_error(f"Cannot index with the token", token)
        return cpl

    def arg_tokens_to_cpl(self, ctx: TranspilerContext, tokens: List[Token]):
        return list(map(lambda x: self.chains_to_cpl(ctx, chain_tokens(x)), split_tokens(tokens, ",")))

    def token_to_cpl(self, ctx: TranspilerContext, t: Token):
        if t.type == TokenType.INT_LITERAL:
            if abs(int(t.value)) > INT_LIMIT:
                raise_syntax_error(
                    f"An integer literal cannot be larger than {INT_LIMIT}", t
                )
            return CplInt(t, t.value)
        if t.type == TokenType.FLOAT_LITERAL:
            if abs(float(t.value)) > get_float_limit():
                raise_syntax_error(
                    f"A float literal cannot be larger than {get_float_limit()}", t
                )
            return CplFloat(t, t.value)
        if t.type == TokenType.STRING_LITERAL:
            return CplString(t, t.value[1:-1])
        if t.type == TokenType.SELECTOR:
            return CplSelector(t, t.value)
        if (
                t.type == TokenType.SELECTOR_IDENTIFIER
                or t.type == TokenType.BLOCK_IDENTIFIER
                or t.type == TokenType.IDENTIFIER
        ):
            return self.variable_token_to_cpl(ctx, t)
        if isinstance(t, GroupToken) and t.func:
            return self.run_function(ctx, t)
        if isinstance(t, GroupToken) and t.open.value == "(":
            return t.cpl(ctx)
        if isinstance(t, GroupToken) and t.open.value == "[":
            if len(t.children) == 0:
                return CplTuple(t, [])
            cpl_list = list(map(
                lambda tk_list: self.tokens_to_cpl(ctx, tk_list), split_tokens(t.children)
            ))
            for cpl in cpl_list:
                if cpl.get_py_value() is None:
                    break
                if cpl.unique_type != cpl_list[0].unique_type:
                    return CplTuple(t, cpl_list)
            return CplArray(t, cpl_list)
        if isinstance(t, GroupToken) and t.open.value == "{":
            d = dict()

            for tl in split_tokens(list(filter(lambda x: x.type not in ENDERS, t.children))):
                if len(tl) < 3:
                    raise_syntax_error("Invalid object", t)
                key_token = tl[0]
                colon_token = tl[1]
                if colon_token.value != ":":
                    raise_syntax_error("Expected a colon", colon_token)
                expr = tl[2:]
                if key_token.type != TokenType.STRING_LITERAL:
                    raise_syntax_error("Expected an object key", key_token)
                    return CplInt(t, 0)
                k = key_token.value[1:-1]
                d[k] = self.tokens_to_cpl(ctx, expr)

            return CplObject(t, d, class_name=None)
        if isinstance(t, LambdaFunctionToken):
            name = f"__lambda__/{get_uuid()}"
            self._transpile_statement(ctx, DefineFunctionStatement(
                t.code,
                t.start,
                t.end,
                Token(name, TokenType.IDENTIFIER, 0, len(name)),
                parse(t.body.children, self.macros, list(self.classes.keys())),
                list(map(lambda x: chain_tokens(x), split_tokens(t.arguments.children, ","))),
                "auto"
            ))
            return CplString(t, name, is_fn_reference=True)
        raise_syntax_error("Unexpected token", t)
        raise ValueError("")

    def chains_to_cpl(
            self, ctx: TranspilerContext, chains: List[List[Token]]
    ) -> CompileTimeValue:
        if len(chains) == 0:
            return CplInt(None, 0)
        if len(chains) == 1:
            return self.chain_to_cpl(ctx, chains[0])
        t0 = chains[0]
        t1 = chains[1]
        if t0[0].value in COMMANDS:
            # TODO: use tokens
            return self.run_cmd(ctx, Token(t0[0].code, TokenType.POINTER, t0[0].start, chains[-1][-1].end))
        if t0[0].value in {"not", "!"}:
            # Basically, 1 - x
            variable_cpl = self.chain_to_cpl(ctx, t1)
            if variable_cpl.unique_type.type not in {"int", "float"}:
                return CplInt(t1[0], 0)
            if isinstance(variable_cpl, CplInt) or isinstance(variable_cpl, CplFloat):
                return CplInt(t1[0], 1 - variable_cpl.value)

            temp_score = CplScore(t1[0], f"int_{get_uuid()}")
            temp_score.compute(ctx, "=", CplInt(None, 1), t1[0])
            variable_cpl.compute(ctx, "-=", temp_score, t1[0])

            return temp_score
        if t0[0].value in INC_OP:
            variable_cpl = self.chain_to_cpl(ctx, t1)

            variable_cpl.compute(ctx, t0[0].value[0] + "=", CplInt(None, 1), t1[0])

            return variable_cpl
        if t1[0].value in INC_OP:
            variable_cpl = self.chain_to_cpl(ctx, t0)

            cached = variable_cpl.cache(ctx)
            variable_cpl.compute(ctx, t1[0].value[0] + "=", CplInt(None, 1), t0[0])

            return cached
        constant_keyword = False
        if t0[0].value == "const":
            constant_keyword = True
            t0 = t1
            if len(chains) < 3:
                raise_syntax_error("Expected an expression for the assignment", t1[0])
            t1 = chains[2]
            chains = chains[1:]

        front_type = cpl_def_from_tokens(self.classes, t0)
        if front_type is not None:
            t0 = t1
            if len(chains) < 3:
                raise_syntax_error("Expected an expression for the assignment", t1[0])
            t1 = chains[2]
            chains = chains[1:]

        if t1[0].value in SET_OP:
            is_constant = constant_keyword
            # TODO: [a, b, c] = [1, 2, 3] syntax and also for objects
            if (
                    t0[0].type != TokenType.IDENTIFIER
                    and t0[0].type != TokenType.SELECTOR_IDENTIFIER
                    and t0[0].type != TokenType.BLOCK_IDENTIFIER
            ):
                raise_syntax_error("Invalid variable name", t0[0])
                raise ValueError("")
            if len(chains) < 3:
                raise_syntax_error("Expected an expression for the assignment", t1[0])
            cpl = self.chains_to_cpl(ctx, chains[2:])
            if t0[0].value == "this" and not ctx.class_name:
                raise_syntax_error("Cannot access 'this' keyword outside of a class", t0[0])
            variable = self._var_to_cpl_or_none(ctx, t0[0])

            var_name_token = t0[0]
            var_name = var_name_token.value
            if isinstance(var_name_token, SelectorIdentifierToken) or isinstance(var_name_token, BlockIdentifierToken):
                var_name = var_name_token.name.value

            is_init = variable is None

            if is_init:
                if t1[0].value != "=" or len(t0) != 1:
                    raise_syntax_error(f"Variable '{t0[0].value}' was not found", t0[0])
                if t0[0] == "this":
                    raise_syntax_error("Cannot assign 'this' keyword to something", t0[0])

                if cpl.unique_type.type in {"int", "float"}:
                    self.load_file.append(
                        f'scoreboard objectives add {var_name} dummy "{var_name}"'
                    )
                if isinstance(cpl, CplTuple) and (not front_type or len(cpl.value) > 0):
                    if len(cpl.value) == 0:
                        show_warning(f"Did you mean to initialize an array instead of a tuple? "
                                     f"Use this syntax: int[] <variable_name> = []", t1[0])
                    is_constant = True
                last_var_type = cpl
                if front_type and front_type != cpl.unique_type:
                    if isinstance(front_type, CplDefArray) and isinstance(cpl, CplTuple) and len(cpl.value) == 0:
                        last_var_type = _type_to_cpl(t0[0], front_type, "", f"storage variables {var_name}")
                    else:
                        raise_syntax_error(f"Variable was defined as a {front_type} but got a {cpl.unique_type}", t0[0])
                if isinstance(cpl, CplSelector):
                    last_var_type = INT_TYPE
                self.variables[var_name] = VariableDeclaration(var_name, last_var_type, is_constant)
                if is_constant:
                    if (isinstance(var_name_token, SelectorIdentifierToken)
                            or isinstance(var_name_token, BlockIdentifierToken)):
                        nbt_loc = self._nbt_var_loc(ctx, var_name_token)
                        ctx.file.append(f"data modify {nbt_loc} set value {json.dumps(cpl.get_py_value())}")
                        del self.variables[var_name]
                    return cpl
            else:
                if front_type:
                    raise_syntax_error("Already defined variables cannot be assigned to a type", t1[0])
                if var_name in self.variables and self.variables[var_name].constant:
                    raise_syntax_error("Variable is constant", t0[0])
                    raise ValueError("")
                if is_constant:
                    raise_syntax_error("Already defined variables cannot be assigned with the 'const' keyword", t0[0])

            variable_cpl = self.chain_to_cpl(ctx, t0)

            if variable_cpl.unique_type.type in {"int", "float"} and isinstance(cpl, CplSelector):
                variable_cpl._set(ctx, CplInt(t0[0], 0))
                if isinstance(variable_cpl, CplScore):
                    set_1 = f"scoreboard players set {variable_cpl.location} 1"
                elif isinstance(variable_cpl, CplNBT):
                    set_1 = f"data modify storage variables {variable_cpl.location} set value 1"
                else:
                    raise ValueError("")
                ctx.file.append(f"execute if entity {cpl.value} run {set_1}")
                return variable_cpl

            if variable_cpl.unique_type == "int" and cpl.unique_type == "float":
                show_warning("Losing float precision by assigning a float to an int.", t1[0])

            variable_cpl.compute(ctx, t1[0].value, cpl, t0[0])

            return cpl

        expr = make_expr(chains)

        stack: List[CompileTimeValue] = []

        for chain in expr:
            if chain[0].type != TokenType.OPERATOR:
                stack.append(self.chain_to_cpl(ctx, chain))
                continue

            right = stack.pop()
            if len(stack) == 0:
                raise_syntax_error("Expected a value after an operator", chain[0])
            left = stack.pop()

            score = left.compute(ctx, chain[0].value, right, t0[0])

            stack.append(score)

        return stack[-1]

    def tokens_to_cpl(
            self, ctx: TranspilerContext, tokens: List[Token]
    ) -> CompileTimeValue:
        if len(tokens) == 0:
            return CplInt(None, 0)
        chains = chain_tokens(tokens)
        return self.chains_to_cpl(ctx, chains)

    def run_function_with_cpl(
            self,
            ctx: TranspilerContext,
            name: str,
            args: List[CompileTimeValue],
            base: Union[Token, None] = None,
            class_name: str | None = None
    ) -> CompileTimeValue:
        exists_fn = self.fn_exists(name)

        if not exists_fn and name not in builtin_fns:
            for var in builtin_vars:
                val = builtin_vars[var].value
                if isinstance(val, CustomCplObject):
                    if name in val.obj:
                        raise_syntax_error(f"Undefined function. Did you mean to use built-in function {var}.{name}()?",
                                           base)
            raise_syntax_error(f"Undefined function: {name}", base)
            return CplInt(base, 0)

        if name not in self.classes and not exists_fn and name in builtin_fns:
            built = builtin_fns[name]
            if len(built) == 1 and built[0].type == "python-raw":
                if not base:
                    raise_syntax_error("Cannot run a raw function without tokens", base)
                if not isinstance(base, GroupToken):
                    raise_syntax_error("Cannot run a raw function without group token", base)
                    raise ValueError("")
                return built[0].function(ctx, raw_group_args(base), base) or built[0].returns
            if len(built) == 1 and built[0].type == "python-cpl":
                return built[0].function(ctx, args, base)
            self.functions.extend(built)

        if exists_fn and exists_fn.type == "python-cpl-imported":
            try:
                if exists_fn.function.__code__.co_argcount == 2:
                    return exists_fn.function(ctx, args) or CplInt(base, 0)
                return exists_fn.function(ctx, args, base)
            except RadonError as e:
                raise e
            except Exception as _:
                stack_trace = traceback.format_exc()
                raise Exception(stack_trace)

        found_fn = None
        available = []
        for f in self.functions:
            if f.name == name:
                available.append(f"{f.name}({', '.join(str(arg.unique_type) + ' ' + arg.name for arg in f.arguments)})")
                if self.check_args(f.arguments, args, base):
                    found_fn = f
                    break
        if found_fn is None:
            raise_syntax_error(f"Invalid arguments. Available usages:\n{'\n'.join(available)}", base)
            assert False

        returns = found_fn.returns
        fn_args = found_fn.arguments

        if found_fn.type == "python-cpl":
            return found_fn.function(ctx, args, base)

        has_any_macro_argument = False

        for index, arg in enumerate(fn_args):
            store_at = arg.store
            val = args[index]
            if arg.store_via == "stack":
                if isinstance(val, CplScore):
                    ctx.file.append(
                        f"data modify {store_at.location} append value {store_at.unique_type.content.get_sample_value()}")
                    val.cache(ctx, nbt_loc=f"{store_at.location}[-1]", force="nbt")
                else:
                    ctx.file.append(f"data modify {store_at.location} append {val.get_data_str(ctx)}")
            elif arg.store_via == "macro":
                val.cache(ctx, nbt_loc=f"storage fn_args_macro {arg.store.location.split(' ')[-1]}", force="nbt")
                has_any_macro_argument = True
            else:
                store_at._set(ctx, val)

        if found_fn.type == "mcfunction-imported":
            eid = f"int_{get_uuid()} __temp__"
            ctx.file.append(
                f"execute store result score {eid} run function {self.pack_namespace}:{found_fn.file_name} with storage fn_mem")
            return CplScore(base, eid)

        if found_fn.type == "python":
            found_fn.function(ctx, found_fn, base)

        actually_returning = CplInt(base, 0)

        if found_fn.type == "radon":
            ctx.file.append(f"function {self.pack_namespace}:{found_fn.file_name}" + (
                f" with storage fn_args_macro" if has_any_macro_argument else ""
            ))
            for index, arg in enumerate(fn_args):
                if arg.store_via == "stack":
                    ctx.file.append(f"data remove {arg.store.location}[-1]")

            # the line after these comments is for the case where you call a function inside a function
            # example:
            # function test(): void {
            #   # "scoreboard players set __returned__ __temp__ 0" runs:
            #   # __returned__ = 0
            #   a()
            #   # __returned__ = 1
            #   # "scoreboard players set __returned__ __temp__ 0" runs again and:
            #   # __returned__ = 0
            #   return
            # }
            # function a(): void {
            #   test()
            #   return
            # }
            if ctx.function and returns != "void":
                ctx.file.append(f"scoreboard players set __returned__ __temp__ 0")
            if isinstance(returns, str):
                if returns == "auto":
                    raise_syntax_error(
                        "Function's return type hasn't been determined yet", base
                    )
            else:
                actually_returning = returns

        if name in self.classes:
            class_name = name

        if class_name is not None and ctx.file is not self.main_file:
            ctx.file.append(f"data modify storage temp class_this set from storage variables this[-1]")
            ctx.file.append(f"data remove storage variables this[-1]")
            actually_returning = CplObjectNBT(base, f"storage temp class_this",
                                              self.classes[class_name].sample.unique_type)

        return actually_returning

    def run_function(
            self, ctx: TranspilerContext, t: GroupToken
    ) -> CompileTimeValue:
        if not t.func:
            return CplInt(t, 0)

        raw_args = []

        if t.func.value in builtin_fns:
            built = builtin_fns[t.func.value]
            if len(built) == 1 and built[0].type == "python-raw":
                return built[0].function(ctx, raw_group_args(t), t) or built[0].returns
            if len(built) == 1 and built[0].type == "python-cpl":
                raw_args = built[0].raw_args

        separated = split_tokens(t.children, ",")
        arguments: List[CompileTimeValue] = []
        for index, arg in enumerate(separated):
            arguments.append(arg if index in raw_args else self.tokens_to_cpl(ctx, arg))

        return self.run_function_with_cpl(ctx, t.func.value, arguments, base=t)

    def compute_tokens(self, ctx: TranspilerContext, left: List[Token], op: str, right: List[Token]):
        return self.tokens_to_cpl(ctx, left).compute(ctx, op, self.tokens_to_cpl(ctx, right), left[0])

    def _cmp_to_cpl(self, ctx, ls):
        return self.compute_tokens(ctx, ls[0], ls[1].value, ls[2])


def _return_safe(ctx: TranspilerContext):
    if ctx.function:
        ctx.file.append(
            f"execute if score __returned__ __temp__ matches 1..1 run return 0"
        )
    if ctx.loop:
        if ctx.file is ctx.loop.file:
            ctx.file.append(
                f"execute if score __break__{ctx.loop.eid} __temp__ matches 1..1 run return 0"
            )
            cou = ctx.loop.continue_
            mac = ""
            if cou[0] == "$":
                mac = "$"
                cou = cou[1:]
            ctx.file.append(
                f"{mac}execute if score __continue__{ctx.loop.eid} __temp__ matches 1..1 run return run {cou}"
            )
        else:
            ctx.file.append(
                f"execute if score __break__{ctx.loop.eid} __temp__ matches 1..1 run return 0"
            )
            ctx.file.append(
                f"execute if score __continue__{ctx.loop.eid} __temp__ matches 1..1 run return 0"
            )



import json
from types import FunctionType
from typing import Union, Any, List, Dict

from ..error import raise_syntax_error, raise_syntax_error_t
from ..tokenizer import Token, GroupToken
from ..utils import CplDef, get_uuid


def tellraw_dumps(obj):
    if isinstance(obj, FunctionType):
        return obj()
    if isinstance(obj, str) or isinstance(obj, bool):
        return json.dumps(obj)
    if isinstance(obj, list):
        return '[' + ",".join(map(tellraw_dumps, obj)) + ']'
    if isinstance(obj, dict):
        return '{' + ",".join(map(lambda x: f'"{x[0]}":{tellraw_dumps(x[1])}', obj.items())) + '}'
    raise ValueError("Unexpected tellraw value: " + str(obj))


def _call_obj(obj):
    for k in obj:
        if isinstance(obj[k], FunctionType):
            obj[k] = obj[k]()
        if isinstance(obj[k], Dict):
            _call_obj(obj[k])
    return obj


class CompileTimeValue:
    def __init__(self, token):
        # type: (Token | None) -> None
        self.unique_type: CplDef = CplDef("unknown")
        self.token = token

    def get_py_value(self) -> Any:
        return None

    def get_data_str(self, ctx):
        # type: (TranspilerContext) -> str
        raise ValueError("")

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(token={self.token}, type={self.unique_type})"

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        # type: (TranspilerContext, str | None, str | None, str | None, str | None) -> Union[CplScore, CplNBT]
        pass

    def _compute_check(self, res, op, cpl):
        if res is None:
            t1 = self.token
            t2 = cpl.token
            if t1.start > t2.start:
                t1 = t2
                t2 = self.token
            raise_syntax_error_t(
                f"Cannot compute {self.__class__.__name__}({self.unique_type}) {op} {cpl.__class__.__name__}({cpl.unique_type})",
                t1.code, t1.start, t2.end)
            assert False

    def compute(self, ctx, op, cpl, token):
        # type: (TranspilerContext, str, CompileTimeValue, Token) -> CompileTimeValue
        st = self.token
        if not st:
            self.token = token
        if op == "+":
            res = self._add(ctx, cpl)
        elif op == "-":
            res = self._sub(ctx, cpl)
        elif op == "*":
            res = self._mul(ctx, cpl)
        elif op == "/":
            res = self._div(ctx, cpl)
        elif op == "%":
            res = self._mod(ctx, cpl)
        elif op == "**":
            res = self._pow(ctx, cpl)
        elif op == "+=":
            res = self._set_add(ctx, cpl)
        elif op == "-=":
            res = self._set_sub(ctx, cpl)
        elif op == "*=":
            res = self._set_mul(ctx, cpl)
        elif op == "/=":
            res = self._set_div(ctx, cpl)
        elif op == "%=":
            res = self._set_mod(ctx, cpl)
        elif op == "**=":
            res = self._set_pow(ctx, cpl)
        elif op == "=":
            res = self._set(ctx, cpl)
        elif op == "==" or op == "is":
            res = self._eq(ctx, cpl)
        elif op == "!=" or op == "is not":
            res = self._neq(ctx, cpl)
        elif op == ">":
            res = self._gt(ctx, cpl)
        elif op == "<":
            res = self._lt(ctx, cpl)
        elif op == ">=":
            res = self._gte(ctx, cpl)
        elif op == "<=":
            res = self._lte(ctx, cpl)
        elif op == "&&" or op == "and":
            res = self._and(ctx, cpl)
        elif op == "||" or op == "or":
            res = self._or(ctx, cpl)
        else:
            assert False
        self._compute_check(res, op, cpl)
        return res

    def cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        # type: (TranspilerContext, str | None, str | None, str | None, str | None) -> Union[CplScore, CplNBT]
        nbt_got_id = None
        if not nbt_loc and force != "score":
            nbt_got_id = get_uuid()
            nbt_loc = f"storage temp _{nbt_got_id}"
        if not score_loc and force != "nbt":
            t = force_t or self.unique_type.type
            score_loc = f"{t}_{nbt_got_id or get_uuid()} __temp__"
        return self._cache(ctx, score_loc, nbt_loc, force, force_t)

    def _get_index(self, ctx, index):
        # type: (TranspilerContext, CompileTimeValue) -> CompileTimeValue | None
        return None

    def _get_slice(self, ctx, index1, index2, index3, token):
        # type: (TranspilerContext, CompileTimeValue, CompileTimeValue, CompileTimeValue, Token) -> CompileTimeValue | None
        return None

    def _call_index(self, ctx, index, arguments, token):
        # type: (TranspilerContext, str, List[CompileTimeValue], Token) -> CompileTimeValue | int | None
        return None

    def _raw_call_index(self, ctx, index, arguments, token):
        # type: (TranspilerContext, str, List[GroupToken], Token) -> CompileTimeValue | int | None
        return None

    def _call(self, ctx, arguments):
        # type: (TranspilerContext, List[CompileTimeValue]) -> CompileTimeValue | None
        return None

    def _raw_call(self, ctx, arguments, token):
        # type: (TranspilerContext, List[GroupToken], Token) -> CompileTimeValue | None
        return None

    def get_index(self, ctx, index, token=None):
        # type: (TranspilerContext, CompileTimeValue, Token | None) -> CompileTimeValue
        st = self.token
        if not st:
            self.token = token
        r = self._get_index(ctx, index)
        self.token = st
        if r is None:
            ind = index
            if isinstance(ind, CompileTimeValue):
                ind = ind.unique_type
            raise_syntax_error(f"Cannot index into {self.unique_type} with {ind}", self.token)
        return r

    def get_slice(self, ctx, index1, index2, index3, token=None):
        # type: (TranspilerContext, CompileTimeValue, CompileTimeValue, CompileTimeValue, Token | None) -> CompileTimeValue
        st = self.token
        if not st:
            self.token = token
        r = self._get_slice(ctx, index1, index2, index3, token)
        self.token = st
        if r is None:
            ind = ":".join(map(lambda x: str(x.unique_type), [index1, index2, index3]))
            raise_syntax_error(f"Cannot index into {self.unique_type} with [{ind}]", self.token)
        return r

    def call_index(self, ctx, index, arguments, token=None):
        # type: (TranspilerContext, str, List[CompileTimeValue] | GroupToken, Token | None) -> CompileTimeValue
        if isinstance(arguments, GroupToken):
            raw_res = self._raw_call_index(ctx, index, raw_group_args(arguments), arguments)
            if raw_res is not None:
                return raw_res
            arguments = ctx.transpiler.arg_tokens_to_cpl(ctx, arguments.children)

        st = self.token
        if not st:
            self.token = token
        r = self._call_index(ctx, index, arguments, token)
        self.token = st
        if r is None:
            raise_syntax_error(f"Cannot call into {self.unique_type} with {index}", token or self.token)
        if isinstance(r, int):
            raise_syntax_error(f"Expected {r} arguments for {self.unique_type}.{index}()", token or self.token)
        return r

    def call(self, ctx, arguments, token=None):
        # type: (TranspilerContext, List[CompileTimeValue] | GroupToken, Token | None) -> CompileTimeValue
        if isinstance(arguments, GroupToken):
            raw_res = self._raw_call(ctx, raw_group_args(arguments), arguments)
            if raw_res is not None:
                return raw_res
            arguments = ctx.transpiler.arg_tokens_to_cpl(ctx, arguments.children)

        st = self.token
        if not st:
            self.token = token
        r = self._call(ctx, arguments)
        self.token = st
        if r is None:
            raise_syntax_error(f"Cannot call into {self.unique_type}", token or self.token)
        if isinstance(r, int):
            raise_syntax_error(f"Expected {r} arguments for {self.unique_type}()", token or self.token)
        return r

    def _set(self, ctx, cpl):
        return None

    def _set_add(self, ctx, cpl):
        return None

    def _set_sub(self, ctx, cpl):
        return None

    def _set_mul(self, ctx, cpl):
        return None

    def _set_div(self, ctx, cpl):
        return None

    def _set_mod(self, ctx, cpl):
        return None

    def _set_pow(self, ctx, cpl):
        return None

    def _add(self, ctx, cpl):
        return None

    def _sub(self, ctx, cpl):
        return None

    def _mul(self, ctx, cpl):
        return None

    def _div(self, ctx, cpl):
        return None

    def _mod(self, ctx, cpl):
        return None

    def _eq_neq(self, ctx, cpl, is_eq):
        return None

    def _eq(self, ctx, cpl):
        return self._eq_neq(ctx, cpl, True)

    def _neq(self, ctx, cpl):
        return self._eq_neq(ctx, cpl, False)

    def _cmp(self, ctx, cpl, op):
        return None

    def _gt(self, ctx, cpl):
        return self._cmp(ctx, cpl, ">")

    def _lt(self, ctx, cpl):
        return self._cmp(ctx, cpl, "<")

    def _gte(self, ctx, cpl):
        return self._cmp(ctx, cpl, ">=")

    def _lte(self, ctx, cpl):
        return self._cmp(ctx, cpl, "<=")

    def _and(self, ctx, cpl):
        return None

    def _or(self, ctx, cpl):
        return None

    def tellraw_object(self, ctx) -> Dict:
        raise_syntax_error(
            f"Cannot read {self.unique_type} value with a $str() macro",
            self.token
        )
        return {}

    def tellraw_object_str(self, ctx):
        return tellraw_dumps(self.tellraw_object(ctx))

    def __iadd__(self, other):
        self._set_add(other[0], other[1])
        return self

    def __isub__(self, other):
        self._set_sub(other[0], other[1])
        return self

    def __imul__(self, other):
        self._set_mul(other[0], other[1])
        return self

    def __idiv__(self, other):
        self._set_div(other[0], other[1])
        return self

    def __imod__(self, other):
        self._set_mod(other[0], other[1])
        return self

    def __add__(self, other):
        return self._add(other[0], other[1])

    def __sub__(self, other):
        return self._sub(other[0], other[1])

    def __mul__(self, other):
        return self._mul(other[0], other[1])

    def __div__(self, other):
        return self._div(other[0], other[1])

    def __mod__(self, other):
        return self._mod(other[0], other[1])

    def __eq__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return self is other
        return self._eq(other[0], other[1])

    def __ne__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return self is not other
        return self._neq(other[0], other[1])

    def __gt__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return False
        return self._gt(other[0], other[1])

    def __lt__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return False
        return self._lt(other[0], other[1])

    def __ge__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return False
        return self._gte(other[0], other[1])

    def __le__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return False
        return self._lte(other[0], other[1])

    def __and__(self, other):
        if not isinstance(other, list) or len(other) != 2 or not isinstance(other[1], CompileTimeValue):
            return False
        return self._and(other[0], other[1])

    def __call__(self, *args, **kwargs):
        return self.call(args[0], list(args[1:]))

    def is_lit_eq(self, v):
        return (isinstance(self, CplInt) or isinstance(self, CplFloat)) and self.value == v


from .nbt import CplNBT
from .int import CplInt
from .float import CplFloat
from .score import CplScore
from ..transpiler import TranspilerContext, raw_group_args



from math import sqrt
from typing import Union, List

from ._literal import CplLiteral
from ._base import CompileTimeValue
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import FLOAT_TYPE, FLOAT_PREC, TokenType


class CplFloat(CplLiteral):
    def __init__(self, token: Union[Token, None], value: float):
        value = float(value)
        if token is None:
            token = Token(str(value), TokenType.FLOAT_LITERAL, 0, len(str(value)))
        super().__init__(token, value)
        self.unique_type = FLOAT_TYPE

    def get_py_value(self):
        return self.value

    def as_int(self):
        return CplInt(self.token, self.value)

    def as_string(self):
        return CplString(self.token, self.value)

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        t = force_t or "float"
        v = self.value
        if force_t == "int":
            v = int(self.value)
        if force == "nbt":
            ctx.file.append(f"data modify {nbt_loc} set value {v}")
            return CplFloatNBT(self.token, nbt_loc)
        if t == "float":
            v = int(v * FLOAT_PREC)
        ctx.file.append(f"scoreboard players set {score_loc} {v}")
        return CplScore(self.token, score_loc)

    def get_data_str(self, ctx):
        return f"value {self.value}"

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        if index == "toString":
            if len(arguments) > 0:
                return 0
            return self.as_string()
        if index == "sqrt":
            return CplFloat(self.token, sqrt(self.value))
        if index == "cbrt":
            return CplFloat(self.token, self.value ** (1 / 3))
        if index == "int":
            if len(arguments) > 0:
                raise_syntax_error("Expected 0 arguments for <float>.int()", self.token)
                assert False
            return self.as_int()
        if index == "float":
            if len(arguments) > 0:
                raise_syntax_error("Expected 0 arguments for <float>.float()", self.token)
                assert False
            return self

    def tellraw_object(self, ctx):
        return {"text": str(self.value)}


from .int import CplInt
from .nbtfloat import CplFloatNBT
from .score import CplScore
from .string import CplString
from ._number import make_num

make_num(CplFloat)
